"""Test OOB Manager."""

import base64
import json
from copy import deepcopy
from datetime import datetime, timedelta, timezone
from typing import List
from unittest import IsolatedAsyncioTestCase
from unittest.mock import ANY

from .....connections.models.conn_record import ConnRecord
from .....connections.models.connection_target import ConnectionTarget
from .....connections.models.diddoc import DIDDoc, PublicKey, PublicKeyType, Service
from .....core.event_bus import EventBus
from .....core.oob_processor import OobMessageProcessor
from .....did.did_key import DIDKey
from .....messaging.decorators.attach_decorator import AttachDecorator
from .....messaging.decorators.service_decorator import ServiceDecorator
from .....messaging.responder import BaseResponder, MockResponder
from .....messaging.util import datetime_now, datetime_to_str, str_to_epoch
from .....multitenant.base import BaseMultitenantManager
from .....multitenant.manager import MultitenantManager
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....transport.inbound.receipt import MessageReceipt
from .....utils.testing import create_test_profile
from .....wallet.askar import AskarWallet
from .....wallet.did_info import DIDInfo, KeyInfo
from .....wallet.did_method import SOV
from .....wallet.key_type import ED25519
from .....wallet.util import pad
from ....coordinate_mediation.v1_0.manager import MediationManager
from ....coordinate_mediation.v1_0.models.mediation_record import MediationRecord
from ....coordinate_mediation.v1_0.route_manager import RouteManager
from ....didcomm_prefix import DIDCommPrefix
from ....didexchange.v1_0.manager import DIDXManager
from ....issue_credential.v1_0.message_types import CREDENTIAL_OFFER
from ....issue_credential.v1_0.messages.credential_offer import (
    CredentialOffer as V10CredOffer,
)
from ....issue_credential.v1_0.messages.inner.credential_preview import (
    CredAttrSpec as V10CredAttrSpec,
)
from ....issue_credential.v1_0.messages.inner.credential_preview import (
    CredentialPreview as V10CredentialPreview,
)
from ....issue_credential.v1_0.models.credential_exchange import V10CredentialExchange
from ....issue_credential.v1_0.tests import INDY_OFFER
from ....issue_credential.v2_0.message_types import (
    ATTACHMENT_FORMAT as V20_CRED_ATTACH_FORMAT,
)
from ....issue_credential.v2_0.message_types import CRED_20_OFFER
from ....issue_credential.v2_0.messages.cred_format import V20CredFormat
from ....issue_credential.v2_0.messages.cred_offer import V20CredOffer
from ....issue_credential.v2_0.messages.inner.cred_preview import (
    V20CredAttrSpec,
    V20CredPreview,
)
from ....present_proof.v1_0.message_types import ATTACH_DECO_IDS as V10_PRES_ATTACH_FORMAT
from ....present_proof.v1_0.message_types import PRESENTATION_REQUEST
from ....present_proof.v1_0.messages.presentation_request import PresentationRequest
from ....present_proof.v2_0.message_types import (
    ATTACHMENT_FORMAT as V20_PRES_ATTACH_FORMAT,
)
from ....present_proof.v2_0.message_types import PRES_20_REQUEST
from ....present_proof.v2_0.messages.pres_format import V20PresFormat
from ....present_proof.v2_0.messages.pres_request import V20PresRequest
from .. import manager as test_module
from ..manager import (
    REUSE_ACCEPTED_WEBHOOK_TOPIC,
    REUSE_WEBHOOK_TOPIC,
    OutOfBandManager,
    OutOfBandManagerError,
)
from ..messages.invitation import HSProto, InvitationMessage
from ..messages.invitation import Service as OobService
from ..messages.problem_report import ProblemReport, ProblemReportReason
from ..messages.reuse import HandshakeReuse
from ..messages.reuse_accept import HandshakeReuseAccept
from ..models.invitation import InvitationRecord
from ..models.oob_record import OobRecord


class TestConfig:
    test_did = "55GkHamhTU1ZbTbV2ab9DE"
    test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
    test_endpoint = "http://localhost"
    test_target_did = "GbuDUYXaUZRfHD2jeDuQuP"
    their_public_did = "55GkHamhTU1ZbTbV2ab9DE"
    test_service = OobService(
        recipient_keys=[test_verkey],
        routing_keys=[],
        service_endpoint=test_endpoint,
    )
    NOW_8601 = datetime.now(tz=timezone.utc).isoformat(" ", "seconds")
    TEST_INVI_MESSAGE_TYPE = "out-of-band/1.1/invitation"
    NOW_EPOCH = str_to_epoch(NOW_8601)
    CD_ID = "GMm4vMw8LLrLJjp81kRRLp:3:CL:12:tag"
    INDY_PROOF_REQ = json.loads(
        f"""{{
        "name": "proof-req",
        "version": "1.0",
        "nonce": "12345",
        "requested_attributes": {{
            "0_player_uuid": {{
                "name": "player",
                "restrictions": [
                    {{
                        "cred_def_id": "{CD_ID}"
                    }}
                ],
                "non_revoked": {{
                    "from": {NOW_EPOCH},
                    "to": {NOW_EPOCH}
                }}
            }},
            "0_screencapture_uuid": {{
                "name": "screenCapture",
                "restrictions": [
                    {{
                        "cred_def_id": "{CD_ID}"
                    }}
                ],
                "non_revoked": {{
                    "from": {NOW_EPOCH},
                    "to": {NOW_EPOCH}
                }}
            }}
        }},
        "requested_predicates": {{
            "0_highscore_GE_uuid": {{
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [
                    {{
                        "cred_def_id": "{CD_ID}"
                    }}
                ],
                "non_revoked": {{
                    "from": {NOW_EPOCH},
                    "to": {NOW_EPOCH}
                }}
            }}
        }}
    }}"""
    )
    DIF_PROOF_REQ = {
        "presentation_definition": {
            "id": "32f54163-7166-48f1-93d8-ff217bdb0654",
            "submission_requirements": [
                {
                    "name": "Citizenship Information",
                    "rule": "pick",
                    "min": 1,
                    "from": "A",
                }
            ],
            "input_descriptors": [
                {
                    "id": "citizenship_input_1",
                    "name": "EU Driver's License",
                    "group": ["A"],
                    "schema": [
                        {
                            "uri": "https://www.w3.org/2018/credentials#VerifiableCredential"
                        }
                    ],
                    "constraints": {
                        "limit_disclosure": "required",
                        "fields": [
                            {
                                "path": ["$.credentialSubject.givenName"],
                                "purpose": "The claim must be from one of the specified issuers",
                                "filter": {
                                    "type": "string",
                                    "enum": ["JOHN", "CAI"],
                                },
                            }
                        ],
                    },
                }
            ],
        },
    }

    PRES_REQ_V1 = PresentationRequest(
        comment="Test",
        request_presentations_attach=[
            AttachDecorator.data_base64(
                mapping=INDY_PROOF_REQ,
                ident=V10_PRES_ATTACH_FORMAT[PRESENTATION_REQUEST],
            )
        ],
    )
    pres_req_dict = PRES_REQ_V1.request_presentations_attach[0].serialize()
    req_attach_v1 = {
        "@id": "request-0",
        "mime-type": "application/json",
        "data": {
            "json": {
                "@type": DIDCommPrefix.qualify_current(PRESENTATION_REQUEST),
                "@id": "12345678-0123-4567-1234-567812345678",
                "comment": "some comment",
                "request_presentations~attach": [pres_req_dict],
            }
        },
    }

    PRES_REQ_V2 = V20PresRequest(
        comment="some comment",
        will_confirm=True,
        formats=[
            V20PresFormat(
                attach_id="indy",
                format_=V20_PRES_ATTACH_FORMAT[PRES_20_REQUEST][
                    V20PresFormat.Format.INDY.api
                ],
            )
        ],
        request_presentations_attach=[
            AttachDecorator.data_base64(mapping=INDY_PROOF_REQ, ident="indy")
        ],
    )

    DIF_PRES_REQ_V2 = V20PresRequest(
        comment="some comment",
        will_confirm=True,
        formats=[
            V20PresFormat(
                attach_id="dif",
                format_=V20_PRES_ATTACH_FORMAT[PRES_20_REQUEST][
                    V20PresFormat.Format.DIF.api
                ],
            )
        ],
        request_presentations_attach=[
            AttachDecorator.data_json(mapping=DIF_PROOF_REQ, ident="dif")
        ],
    )

    CRED_OFFER_V1 = V10CredOffer(
        credential_preview=V10CredentialPreview(
            attributes=(
                V10CredAttrSpec(name="legalName", value="value"),
                V10CredAttrSpec(name="jurisdictionId", value="value"),
                V10CredAttrSpec(name="incorporationDate", value="value"),
            )
        ),
        offers_attach=[V10CredOffer.wrap_indy_offer(INDY_OFFER)],
    )

    CRED_OFFER_V2 = V20CredOffer(
        credential_preview=V20CredPreview(
            attributes=V20CredAttrSpec.list_plain(
                {
                    "legalName": "value",
                    "jurisdictionId": "value",
                    "incorporationDate": "value",
                }
            ),
        ),
        formats=[
            V20CredFormat(
                attach_id="indy",
                format_=V20_CRED_ATTACH_FORMAT[CRED_20_OFFER][
                    V20CredFormat.Format.INDY.api
                ],
            )
        ],
        offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="indy")],
    )

    req_attach_v2 = AttachDecorator.data_json(
        mapping=PRES_REQ_V2.serialize(),
        ident="request-0",
    ).serialize()

    indy_cred_req = {
        "schema_id": f"{test_did}:2:bc-reg:1.0",
        "cred_def_id": f"{test_did}:3:CL:12:tag1",
    }
    cred_req_meta = {}

    def make_did_doc(self, did, verkey):
        doc = DIDDoc(did=did)
        controller = did
        ident = "1"
        pk_value = verkey
        pk = PublicKey(
            did, ident, pk_value, PublicKeyType.ED25519_SIG_2018, controller, False
        )
        doc.set(pk)
        recip_keys = [pk]
        router_keys = []
        service = Service(
            did, "indy", "IndyAgent", recip_keys, router_keys, TestConfig.test_endpoint
        )
        doc.set(service)
        return doc


class TestOOBManager(IsolatedAsyncioTestCase, TestConfig):
    async def asyncSetUp(self):
        self.responder = MockResponder()
        self.responder.send = mock.CoroutineMock()

        self.test_mediator_routing_keys = ["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRR"]
        self.test_mediator_conn_id = "mediator-conn-id"
        self.test_mediator_endpoint = "http://mediator.example.com"

        self.route_manager = mock.MagicMock(RouteManager, autospec=True)
        self.route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )

        self.profile = await create_test_profile(
            {
                "default_endpoint": TestConfig.test_endpoint,
                "default_label": "This guy",
                "additional_endpoints": ["http://aries.ca/another-endpoint"],
                "debug.auto_accept_invites": True,
                "debug.auto_accept_requests": True,
            }
        )

        self.profile.context.injector.bind_instance(RouteManager, self.route_manager)

        self.profile.context.injector.bind_instance(BaseResponder, self.responder)

        self.profile.context.injector.bind_instance(
            EventBus, mock.MagicMock(EventBus, autospec=True)
        )
        self.multitenant_mgr = mock.MagicMock(MultitenantManager, autospec=True)
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, self.multitenant_mgr
        )

        self.manager = OutOfBandManager(self.profile)
        assert self.manager.profile
        self.manager.resolve_invitation = mock.CoroutineMock()
        self.manager.resolve_invitation.return_value = (
            TestConfig.test_endpoint,
            [TestConfig.test_verkey],
            [],
        )

        self.test_conn_rec = mock.MagicMock(
            connection_id="dummy",
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_role=ConnRecord.Role.REQUESTER,
            state=ConnRecord.State.COMPLETED,
            their_public_did=self.their_public_did,
            save=mock.CoroutineMock(),
        )

    async def test_create_invitation_handshake_succeeds(self):
        self.profile.context.update_settings({"public_invites": True})

        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=[HSProto.RFC23],
            )

            assert invi_rec.invitation._type == DIDCommPrefix.qualify_current(
                self.TEST_INVI_MESSAGE_TYPE
            )
            assert not invi_rec.invitation.requests_attach
            assert (
                DIDCommPrefix.qualify_current(HSProto.RFC23.name)
                in invi_rec.invitation.handshake_protocols
            )
            assert invi_rec.invitation.services == [f"did:sov:{TestConfig.test_did}"]

    async def test_create_invitation_multitenant_local(self):
        self.profile.context.update_settings(
            {
                "multitenant.enabled": True,
                "wallet.id": "test_wallet",
            }
        )

        with (
            mock.patch.object(
                AskarWallet, "create_signing_key", autospec=True
            ) as mock_wallet_create_signing_key,
            mock.patch.object(
                self.multitenant_mgr, "get_default_mediator"
            ) as mock_get_default_mediator,
        ):
            mock_wallet_create_signing_key.return_value = KeyInfo(
                TestConfig.test_verkey, None, ED25519
            )
            mock_get_default_mediator.return_value = MediationRecord()
            await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                hs_protos=[HSProto.RFC23],
                multi_use=False,
            )

            self.route_manager.route_invitation.assert_called_once()

    async def test_create_invitation_multitenant_public(self):
        self.profile.context.update_settings(
            {
                "multitenant.enabled": True,
                "wallet.id": "test_wallet",
                "public_invites": True,
            }
        )

        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            await self.manager.create_invitation(
                hs_protos=[HSProto.RFC23],
                public=True,
                multi_use=False,
            )

            self.route_manager.route_invitation.assert_called_once()

    async def test_create_invitation_mediation_overwrites_routing_and_endpoint(self):
        async with self.profile.session() as session:
            mock_conn_rec = mock.MagicMock()

            mediation_record = MediationRecord(
                role=MediationRecord.ROLE_CLIENT,
                state=MediationRecord.STATE_GRANTED,
                connection_id=self.test_mediator_conn_id,
                routing_keys=self.test_mediator_routing_keys,
                endpoint=self.test_mediator_endpoint,
            )
            await mediation_record.save(session)
            with (
                mock.patch.object(
                    MediationManager,
                    "get_default_mediator_id",
                ) as mock_get_default_mediator,
                mock.patch.object(mock_conn_rec, "metadata_set", mock.CoroutineMock()),
            ):
                invite = await self.manager.create_invitation(
                    my_endpoint=TestConfig.test_endpoint,
                    my_label="test123",
                    hs_protos=[HSProto.RFC23],
                    mediation_id=mediation_record.mediation_id,
                )
                assert isinstance(invite, InvitationRecord)
                assert invite.invitation._type == DIDCommPrefix.qualify_current(
                    self.TEST_INVI_MESSAGE_TYPE
                )
                assert invite.invitation.label == "test123"
                assert (
                    DIDKey.from_did(
                        invite.invitation.services[0].routing_keys[0]
                    ).public_key_b58
                    == self.test_mediator_routing_keys[0]
                )
                assert (
                    invite.invitation.services[0].service_endpoint
                    == self.test_mediator_endpoint
                )
                mock_get_default_mediator.assert_not_called()

    async def test_create_invitation_no_handshake_no_attachments_x(self):
        with self.assertRaises(OutOfBandManagerError) as context:
            await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=None,
                multi_use=False,
            )
        assert "Invitation must include" in str(context.exception)

    async def test_create_invitation_attachment_v1_0_cred_offer(self):
        self.profile.context.update_settings({"public_invites": True})
        with (
            mock.patch.object(
                AskarWallet, "get_public_did", autospec=True
            ) as mock_wallet_get_public_did,
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_cxid,
        ):
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            mock_retrieve_cxid.return_value = mock.MagicMock(
                credential_offer_dict=self.CRED_OFFER_V1
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=[HSProto.RFC23],
                multi_use=False,
                attachments=[{"type": "credential-offer", "id": "dummy-id"}],
            )

            mock_retrieve_cxid.assert_called_once_with(ANY, "dummy-id")
            assert isinstance(invi_rec, InvitationRecord)
            assert invi_rec.invitation.handshake_protocols
            assert invi_rec.invitation.requests_attach[0].content[
                "@type"
            ] == DIDCommPrefix.qualify_current(CREDENTIAL_OFFER)

    async def test_create_invitation_attachment_v1_0_cred_offer_no_handshake(self):
        self.profile.context.update_settings({"public_invites": True})
        with (
            mock.patch.object(
                AskarWallet, "get_public_did", autospec=True
            ) as mock_wallet_get_public_did,
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_cxid,
        ):
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            mock_retrieve_cxid.return_value = mock.MagicMock(
                credential_offer_dict=self.CRED_OFFER_V1
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=None,
                multi_use=False,
                attachments=[{"type": "credential-offer", "id": "dummy-id"}],
            )

            mock_retrieve_cxid.assert_called_once_with(ANY, "dummy-id")
            assert isinstance(invi_rec, InvitationRecord)
            assert not invi_rec.invitation.handshake_protocols
            assert invi_rec.invitation.requests_attach[0].content == {
                **self.CRED_OFFER_V1.serialize(),
                "~thread": {"pthid": invi_rec.invi_msg_id},
            }

    async def test_create_invitation_attachment_v2_0_cred_offer(self):
        with (
            mock.patch.object(
                AskarWallet, "get_public_did", autospec=True
            ) as mock_wallet_get_public_did,
            mock.patch.object(
                test_module.V10CredentialExchange,
                "retrieve_by_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_cxid_v1,
            mock.patch.object(
                test_module.V20CredExRecord,
                "retrieve_by_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_cxid_v2,
        ):
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            mock_retrieve_cxid_v1.side_effect = test_module.StorageNotFoundError()
            mock_retrieve_cxid_v2.return_value = mock.MagicMock(cred_offer=V20CredOffer())
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=False,
                hs_protos=None,
                multi_use=False,
                attachments=[{"type": "credential-offer", "id": "dummy-id"}],
            )

            mock_retrieve_cxid_v2.assert_called_once_with(ANY, "dummy-id")
            assert isinstance(invi_rec, InvitationRecord)
            assert not invi_rec.invitation.handshake_protocols
            attach = invi_rec.invitation.requests_attach[0].content
            assert isinstance(attach, dict)
            assert "~thread" in attach and "pthid" in attach["~thread"]
            assert attach["~thread"]["pthid"] == invi_rec.invi_msg_id

    async def test_create_invitation_attachment_present_proof_v1_0(self):
        self.profile.context.update_settings({"public_invites": True})
        with (
            mock.patch.object(
                AskarWallet, "get_public_did", autospec=True
            ) as mock_wallet_get_public_did,
            mock.patch.object(
                test_module.V10PresentationExchange,
                "retrieve_by_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_pxid,
        ):
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            mock_retrieve_pxid.return_value = mock.MagicMock(
                presentation_request_dict=self.PRES_REQ_V1
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=[test_module.HSProto.RFC23],
                multi_use=False,
                attachments=[{"type": "present-proof", "id": "dummy-id"}],
            )

            mock_retrieve_pxid.assert_called_once_with(ANY, "dummy-id")
            assert isinstance(invi_rec, InvitationRecord)
            assert invi_rec.invitation.handshake_protocols
            assert invi_rec.invitation.requests_attach[0].content == {
                **self.PRES_REQ_V1.serialize(),
                "~thread": {"pthid": invi_rec.invi_msg_id},
            }

    async def test_create_invitation_attachment_present_proof_v2_0(self):
        self.profile.context.update_settings({"public_invites": True})
        with (
            mock.patch.object(
                AskarWallet, "get_public_did", autospec=True
            ) as mock_wallet_get_public_did,
            mock.patch.object(
                test_module.V10PresentationExchange,
                "retrieve_by_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_pxid_1,
            mock.patch.object(
                test_module.V20PresExRecord,
                "retrieve_by_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_pxid_2,
        ):
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            mock_retrieve_pxid_1.side_effect = StorageNotFoundError()
            mock_retrieve_pxid_2.return_value = mock.MagicMock(
                pres_request=TestConfig.PRES_REQ_V2
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=[test_module.HSProto.RFC23],
                multi_use=False,
                attachments=[{"type": "present-proof", "id": "dummy-id"}],
            )

            mock_retrieve_pxid_2.assert_called_once_with(ANY, "dummy-id")
            assert isinstance(invi_rec, InvitationRecord)
            assert invi_rec.invitation.handshake_protocols
            assert invi_rec.invitation.requests_attach[0].content == {
                **TestConfig.PRES_REQ_V2.serialize(),
                "~thread": {"pthid": invi_rec.invi_msg_id},
            }

    async def test_create_invitation_public_x_no_public_invites(self):
        self.profile.context.update_settings({"public_invites": False})

        with self.assertRaises(OutOfBandManagerError) as context:
            await self.manager.create_invitation(
                public=True,
                my_endpoint="testendpoint",
                hs_protos=[test_module.HSProto.RFC23],
                multi_use=False,
            )
        assert "Public invitations are not enabled" in str(context.exception)

    async def test_create_invitation_public_x_multi_use(self):
        self.profile.context.update_settings({"public_invites": True})

        with self.assertRaises(OutOfBandManagerError) as context:
            await self.manager.create_invitation(
                public=True,
                my_endpoint="testendpoint",
                hs_protos=[test_module.HSProto.RFC23],
                multi_use=True,
            )
        assert "Cannot create public invitation with" in str(context.exception)

    async def test_create_invitation_requests_attach_x_multi_use(self):
        with self.assertRaises(OutOfBandManagerError) as context:
            await self.manager.create_invitation(
                public=False,
                my_endpoint="testendpoint",
                hs_protos=[test_module.HSProto.RFC23],
                attachments=[{"some": "attachment"}],
                multi_use=True,
            )
        assert "Cannot create multi use invitation with attachments" in str(
            context.exception
        )

    async def test_create_invitation_public_x_no_public_did(self):
        self.profile.context.update_settings({"public_invites": True})

        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = None
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.create_invitation(
                    public=True,
                    my_endpoint="testendpoint",
                    hs_protos=[test_module.HSProto.RFC23],
                    multi_use=False,
                )
            assert "Cannot create public invitation with no public DID" in str(
                context.exception
            )

    async def test_create_invitation_attachment_x(self):
        self.profile.context.update_settings({"public_invites": True})
        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.create_invitation(
                    my_endpoint=TestConfig.test_endpoint,
                    public=False,
                    hs_protos=[test_module.HSProto.RFC23],
                    multi_use=False,
                    attachments=[
                        {
                            "type": "asdf",
                            "id": "asdf",
                            "having": "attachment",
                            "is": "no",
                            "good": "here",
                        }
                    ],
                )
            assert "Unknown attachment type" in str(context.exception)

    async def test_create_invitation_peer_did(self):
        async with self.profile.session() as session:
            self.profile.context.update_settings(
                {
                    "multitenant.enabled": True,
                    "wallet.id": "my-wallet",
                }
            )
            mediation_record = MediationRecord(
                role=MediationRecord.ROLE_CLIENT,
                state=MediationRecord.STATE_GRANTED,
                connection_id=self.test_mediator_conn_id,
                routing_keys=self.test_mediator_routing_keys,
                endpoint=self.test_mediator_endpoint,
            )
            await mediation_record.save(session)
            with mock.patch.object(
                self.multitenant_mgr, "get_default_mediator"
            ) as mock_get_default_mediator:
                mock_get_default_mediator.return_value = mediation_record
                invi_rec = await self.manager.create_invitation(
                    my_label="That guy",
                    my_endpoint=None,
                    public=False,
                    hs_protos=[test_module.HSProto.RFC23],
                    multi_use=False,
                    service_accept=["didcomm/aip1", "didcomm/aip2;env=rfc19"],
                )

                assert invi_rec._invitation.ser["@type"] == DIDCommPrefix.qualify_current(
                    self.TEST_INVI_MESSAGE_TYPE
                )
                assert not invi_rec._invitation.ser.get("requests~attach")
                assert invi_rec.invitation.label == "That guy"
                assert (
                    DIDCommPrefix.qualify_current(HSProto.RFC23.name)
                    in invi_rec.invitation.handshake_protocols
                )
                service = invi_rec._invitation.ser["services"][0]
                assert service["id"] == "#inline"
                assert service["type"] == "did-communication"
                assert len(service["recipientKeys"]) == 1
                assert (
                    service["routingKeys"][0]
                    == DIDKey.from_public_key_b58(
                        self.test_mediator_routing_keys[0], ED25519
                    ).key_id
                )
                assert service["serviceEndpoint"] == self.test_mediator_endpoint

    async def test_create_invitation_metadata_assigned(self):
        async with self.profile.session() as session:
            invi_rec = await self.manager.create_invitation(
                hs_protos=[test_module.HSProto.RFC23],
                metadata={"hello": "world"},
            )
            service = invi_rec._invitation.ser["services"][0]
            invitation_key = DIDKey.from_did(service["recipientKeys"][0]).public_key_b58
            record = await ConnRecord.retrieve_by_invitation_key(session, invitation_key)
            assert await record.metadata_get_all(session) == {"hello": "world"}

    async def test_create_invitation_x_public_metadata(self):
        self.profile.context.update_settings({"public_invites": True})
        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.create_invitation(
                    public=False,
                    hs_protos=[],
                    attachments=[{"an": "attachment"}],
                    metadata={"hello": "world"},
                    multi_use=False,
                )
            assert "Cannot store metadata without handshake protocols" in str(
                context.exception
            )

    async def test_create_invitation_goal_code(self):
        self.profile.context.update_settings({"public_invites": True})
        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            goal = "To issue a Faber College Graduate credential"
            goal_code = "issue-vc"
            my_label = "Invitation to Barry"

            invi_rec = await self.manager.create_invitation(
                public=False,
                my_label=my_label,
                protocol_version="1.1",
                alias="Test Alias",
                service_accept=["didcomm/aip1", "didcomm/aip2;env=rfc19"],
                hs_protos=[test_module.HSProto.RFC23],
                goal=goal,
                goal_code=goal_code,
            )
            assert isinstance(invi_rec, InvitationRecord)
            assert invi_rec.invitation.handshake_protocols
            assert invi_rec.invitation.label == my_label
            assert invi_rec.invitation.goal == goal
            assert invi_rec.invitation.goal_code == goal_code
            assert invi_rec.invitation_url

            base64_message = invi_rec.invitation_url.split("=", maxsplit=1)[1]
            base64_bytes = base64_message.encode("ascii")
            message_bytes = base64.b64decode(base64_bytes + b"==")
            base64.urlsafe_b64decode(pad(base64_message))
            data = message_bytes.decode("ascii")
            assert data
            invite_json = json.loads(data)
            assert invite_json
            assert invite_json["label"] == my_label
            assert invite_json["goal"] == goal
            assert invite_json["goal_code"] == goal_code

    async def test_wait_for_conn_rec_active_retrieve_by_id(self):
        with mock.patch.object(
            ConnRecord,
            "retrieve_by_id",
            mock.CoroutineMock(
                return_value=mock.MagicMock(connection_id="the-retrieved-connection-id")
            ),
        ):
            conn_rec = await self.manager._wait_for_conn_rec_active("a-connection-id")
            assert conn_rec.connection_id == "the-retrieved-connection-id"

    async def test_create_handshake_reuse_msg(self):
        self.profile.context.update_settings({"public_invites": True})

        with (
            mock.patch.object(
                OutOfBandManager,
                "fetch_connection_targets",
                autospec=True,
            ) as oob_mgr_fetch_conn,
            mock.patch.object(
                ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=self.test_conn_rec),
            ),
        ):
            oob_mgr_fetch_conn.return_value = ConnectionTarget(
                did=TestConfig.test_did,
                endpoint=TestConfig.test_endpoint,
                recipient_keys=[TestConfig.test_verkey],
                sender_key=TestConfig.test_verkey,
            )

            invitation = InvitationMessage()
            oob_record = OobRecord(
                invitation=invitation,
                invi_msg_id=invitation._id,
                role=OobRecord.ROLE_RECEIVER,
                connection_id=self.test_conn_rec.connection_id,
                state=OobRecord.STATE_INITIAL,
            )

            oob_record = await self.manager._create_handshake_reuse_message(
                oob_record, self.test_conn_rec, invitation._version
            )

            _, kwargs = self.responder.send.call_args
            reuse_message: HandshakeReuse = kwargs.get("message")

            assert oob_record.state == OobRecord.STATE_AWAIT_RESPONSE

            # Assert responder has been called with the reuse message
            assert reuse_message._type == DIDCommPrefix.qualify_current(
                "out-of-band/1.1/handshake-reuse"
            )
            assert oob_record.reuse_msg_id == reuse_message._id

    async def test_create_handshake_reuse_msg_catch_exception(self):
        self.profile.context.update_settings({"public_invites": True})
        with mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            oob_mgr_fetch_conn.side_effect = StorageNotFoundError()
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager._create_handshake_reuse_message(
                    mock.MagicMock(), self.test_conn_rec, "1.0"
                )
            assert "Error on creating and sending a handshake reuse message" in str(
                context.exception
            )

    async def test_receive_reuse_message_existing_found(self):
        self.profile.context.update_settings({"public_invites": True})

        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
        )

        reuse_msg = HandshakeReuse()
        reuse_msg.assign_thread_id(thid="the-thread-id", pthid="the-pthid")

        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160

        with (
            mock.patch.object(
                OutOfBandManager,
                "fetch_connection_targets",
                autospec=True,
            ) as oob_mgr_fetch_conn,
            mock.patch.object(
                OobRecord,
                "retrieve_by_tag_filter",
                autospec=True,
            ) as mock_retrieve_oob,
            mock.patch.object(self.profile, "notify", autospec=True) as mock_notify,
        ):
            mock_retrieve_oob.return_value = mock.MagicMock(
                emit_event=mock.CoroutineMock(),
                delete_record=mock.CoroutineMock(),
                multi_use=False,
            )

            await self.manager.receive_reuse_message(
                reuse_msg, receipt, self.test_conn_rec
            )
            mock_notify.assert_called_once_with(
                REUSE_WEBHOOK_TOPIC,
                {
                    "thread_id": "the-thread-id",
                    "connection_id": self.test_conn_rec.connection_id,
                    "comment": "Connection dummy is being reused for invitation the-pthid",
                },
            )

            # delete should be called if multi_use == False
            mock_retrieve_oob.return_value.delete_record.assert_called_once()
            mock_retrieve_oob.return_value.emit_event.assert_called_once()
            self.responder.send.assert_called_once_with(
                message=ANY, target_list=oob_mgr_fetch_conn.return_value
            )

            assert mock_retrieve_oob.return_value.state == OobRecord.STATE_DONE
            assert mock_retrieve_oob.return_value.reuse_msg_id == reuse_msg._thread_id
            assert (
                mock_retrieve_oob.return_value.connection_id
                == self.test_conn_rec.connection_id
            )

    async def test_receive_reuse_message_existing_found_multi_use(self):
        self.profile.context.update_settings({"public_invites": True})

        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
        )

        reuse_msg = HandshakeReuse(version="1.0")
        reuse_msg.assign_thread_id(thid="the-thread-id", pthid="the-pthid")

        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160

        with (
            mock.patch.object(
                OutOfBandManager,
                "fetch_connection_targets",
                autospec=True,
            ) as oob_mgr_fetch_conn,
            mock.patch.object(
                OobRecord,
                "retrieve_by_tag_filter",
                autospec=True,
            ) as mock_retrieve_oob,
            mock.patch.object(self.profile, "notify", autospec=True) as mock_notify,
        ):
            mock_retrieve_oob.return_value = mock.MagicMock(
                emit_event=mock.CoroutineMock(),
                delete_record=mock.CoroutineMock(),
                multi_use=True,
            )

            await self.manager.receive_reuse_message(
                reuse_msg, receipt, self.test_conn_rec
            )
            mock_notify.assert_called_once_with(
                REUSE_WEBHOOK_TOPIC,
                {
                    "thread_id": "the-thread-id",
                    "connection_id": self.test_conn_rec.connection_id,
                    "comment": "Connection dummy is being reused for invitation the-pthid",
                },
            )

            # delete should be called if multi_use == False
            mock_retrieve_oob.return_value.delete_record.assert_not_called()
            mock_retrieve_oob.return_value.emit_event.assert_called_once()
            self.responder.send.assert_called_once_with(
                message=ANY, target_list=oob_mgr_fetch_conn.return_value
            )

            assert mock_retrieve_oob.return_value.state == OobRecord.STATE_DONE
            assert mock_retrieve_oob.return_value.reuse_msg_id == reuse_msg._thread_id
            assert (
                mock_retrieve_oob.return_value.connection_id
                == self.test_conn_rec.connection_id
            )

    async def test_receive_reuse_accepted(self):
        self.profile.context.update_settings({"public_invites": True})

        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg_accepted = HandshakeReuseAccept()
        reuse_msg_accepted.assign_thread_id(thid="the-thread-id", pthid="the-pthid")

        with (
            mock.patch.object(self.profile, "notify", autospec=True) as mock_notify,
            mock.patch.object(
                OobRecord, "retrieve_by_tag_filter", autospec=True
            ) as mock_retrieve_oob,
        ):
            mock_retrieve_oob.return_value = mock.MagicMock(
                emit_event=mock.CoroutineMock(),
                delete_record=mock.CoroutineMock(),
            )

            await self.manager.receive_reuse_accepted_message(
                reuse_msg_accepted, receipt, self.test_conn_rec
            )

            mock_retrieve_oob.return_value.emit_event.assert_called_once()
            mock_retrieve_oob.return_value.delete_record.assert_called_once()
            mock_notify.assert_called_once_with(
                REUSE_ACCEPTED_WEBHOOK_TOPIC,
                {
                    "thread_id": "the-thread-id",
                    "connection_id": self.test_conn_rec.connection_id,
                    "state": "accepted",
                    "comment": f"Connection {self.test_conn_rec.connection_id} is being reused for invitation the-pthid",
                },
            )

    async def test_receive_reuse_accepted_x(self):
        self.profile.context.update_settings({"public_invites": True})

        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg_accepted = HandshakeReuseAccept()
        reuse_msg_accepted.assign_thread_id(thid="the-thread-id", pthid="the-pthid")

        with (
            mock.patch.object(self.profile, "notify", autospec=True) as mock_notify,
            mock.patch.object(
                OobRecord, "retrieve_by_tag_filter", autospec=True
            ) as mock_retrieve_oob,
        ):
            mock_retrieve_oob.side_effect = (StorageNotFoundError,)

            with self.assertRaises(test_module.OutOfBandManagerError) as err:
                await self.manager.receive_reuse_accepted_message(
                    reuse_msg_accepted, receipt, self.test_conn_rec
                )
            assert "Error processing reuse accepted message " in err.exception.message

            mock_notify.assert_called_once_with(
                REUSE_ACCEPTED_WEBHOOK_TOPIC,
                {
                    "thread_id": "the-thread-id",
                    "state": "rejected",
                    "connection_id": self.test_conn_rec.connection_id,
                    "comment": f"Unable to process HandshakeReuseAccept message, connection {self.test_conn_rec.connection_id} and invitation the-pthid",
                },
            )

    async def test_receive_problem_report(self):
        self.profile.context.update_settings({"public_invites": True})

        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        problem_report = ProblemReport(
            description={
                "en": "test",
                "code": ProblemReportReason.EXISTING_CONNECTION_NOT_ACTIVE.value,
            }
        )
        problem_report.assign_thread_id(thid="the-thread-id", pthid="the-pthid")

        with mock.patch.object(
            OobRecord, "retrieve_by_tag_filter", autospec=True
        ) as mock_retrieve_oob:
            mock_retrieve_oob.return_value = mock.MagicMock(
                emit_event=mock.CoroutineMock(),
                delete_record=mock.CoroutineMock(),
                save=mock.CoroutineMock(),
            )

            await self.manager.receive_problem_report(
                problem_report, receipt, self.test_conn_rec
            )

            mock_retrieve_oob.assert_called_once_with(
                ANY, {"invi_msg_id": "the-pthid", "reuse_msg_id": "the-thread-id"}
            )
            assert mock_retrieve_oob.return_value.state == OobRecord.STATE_NOT_ACCEPTED

    async def test_receive_problem_report_x(self):
        self.profile.context.update_settings({"public_invites": True})

        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        problem_report = ProblemReport(
            description={
                "en": "test",
                "code": ProblemReportReason.EXISTING_CONNECTION_NOT_ACTIVE.value,
            }
        )
        problem_report.assign_thread_id(thid="the-thread-id", pthid="the-pthid")

        with mock.patch.object(
            OobRecord, "retrieve_by_tag_filter", autospec=True
        ) as mock_retrieve_oob:
            mock_retrieve_oob.side_effect = (StorageNotFoundError(),)

            with self.assertRaises(OutOfBandManagerError) as err:
                await self.manager.receive_problem_report(
                    problem_report, receipt, self.test_conn_rec
                )
            assert "Error processing problem report message " in err.exception.message

    async def test_receive_invitation_with_valid_mediation(self):
        mock_conn = mock.MagicMock(connection_id="dummy-connection")

        async with self.profile.session() as session:
            self.profile.context.update_settings({"public_invites": True})
            mediation_record = MediationRecord(
                role=MediationRecord.ROLE_CLIENT,
                state=MediationRecord.STATE_GRANTED,
                connection_id=self.test_mediator_conn_id,
                routing_keys=self.test_mediator_routing_keys,
                endpoint=self.test_mediator_endpoint,
            )
            await mediation_record.save(session)
            with (
                mock.patch.object(
                    DIDXManager, "receive_invitation", mock.CoroutineMock()
                ) as mock_didx_recv_invi,
                mock.patch.object(
                    ConnRecord, "retrieve_by_id", mock.CoroutineMock()
                ) as mock_retrieve_conn_by_id,
            ):
                invite = await self.manager.create_invitation(
                    my_endpoint=TestConfig.test_endpoint,
                    my_label="test123",
                    hs_protos=[HSProto.RFC23],
                )

                mock_retrieve_conn_by_id.return_value = mock_conn
                mock_didx_recv_invi.return_value = mock_conn
                invi_msg = invite.invitation
                await self.manager.receive_invitation(
                    invitation=invi_msg,
                    mediation_id=mediation_record._id,
                )
                mock_didx_recv_invi.assert_called_once_with(
                    invitation=invi_msg,
                    their_public_did=None,
                    auto_accept=None,
                    alias=None,
                    mediation_id=mediation_record._id,
                    protocol="didexchange/1.0",
                )

    async def test_receive_invitation_with_invalid_mediation(self):
        mock_conn = mock.MagicMock(connection_id="dummy-connection")

        with (
            mock.patch.object(
                DIDXManager,
                "receive_invitation",
                mock.CoroutineMock(),
            ) as mock_didx_recv_invi,
            mock.patch.object(
                ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(),
            ) as mock_retrieve_conn_by_id,
        ):
            invite = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                my_label="test123",
                hs_protos=[HSProto.RFC23],
            )
            mock_didx_recv_invi.return_value = mock_conn
            mock_retrieve_conn_by_id.return_value = mock_conn
            invi_msg = invite.invitation
            self.route_manager.mediation_record_if_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError
            )
            await self.manager.receive_invitation(
                invi_msg,
                mediation_id="test-mediation-id",
            )
            mock_didx_recv_invi.assert_called_once_with(
                invitation=invi_msg,
                their_public_did=None,
                auto_accept=None,
                alias=None,
                mediation_id=None,
                protocol="didexchange/1.0",
            )

    async def test_receive_invitation_didx_services_with_service_block(self):
        self.profile.context.update_settings({"public_invites": True})

        mock_conn = mock.MagicMock(connection_id="dummy-connection")

        with (
            mock.patch.object(test_module, "DIDXManager", autospec=True) as didx_mgr_cls,
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_retrieve_conn_by_id,
        ):
            didx_mgr_cls.return_value = mock.MagicMock(
                receive_invitation=mock.CoroutineMock(return_value=mock_conn)
            )
            mock_retrieve_conn_by_id.return_value = mock_conn
            oob_invitation = InvitationMessage(
                requests_attach=[],
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                services=[
                    OobService(
                        recipient_keys=["dummy"],
                        routing_keys=[],
                    )
                ],
            )

            await self.manager.receive_invitation(oob_invitation)

    async def test_receive_invitation_services_with_neither_service_blocks_nor_dids(
        self,
    ):
        self.profile.context.update_settings({"public_invites": True})
        oob_invitation = InvitationMessage(
            services=[],
        )
        with self.assertRaises(OutOfBandManagerError) as err:
            await self.manager.receive_invitation(oob_invitation)

        assert "service array must have exactly one element" in err.exception.message

    async def test_receive_invitation_no_hs_protos_no_attach(
        self,
    ):
        self.profile.context.update_settings({"public_invites": True})
        oob_invitation = InvitationMessage(
            services=["did:sov:something"],
        )
        with self.assertRaises(OutOfBandManagerError) as err:
            await self.manager.receive_invitation(oob_invitation)

        assert (
            "Invitation must specify handshake_protocols, requests_attach, or both"
            in err.exception.message
        )

    async def test_existing_conn_record_public_did(self):
        self.profile.context.update_settings({"public_invites": True})

        test_exist_conn = ConnRecord(
            connection_id="connection_id",
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-0123-4567-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )

        with (
            mock.patch.object(
                ConnRecord,
                "find_existing_connection",
                mock.CoroutineMock(),
            ) as oob_mgr_find_existing_conn,
            mock.patch.object(OobRecord, "save", mock.CoroutineMock()) as oob_record_save,
            mock.patch.object(
                OobRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as oob_record_retrieve_by_id,
            mock.patch.object(
                OutOfBandManager, "fetch_connection_targets", autospec=True
            ) as oob_mgr_fetch_conn,
        ):
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            oob_mgr_fetch_conn.return_value = []
            oob_invitation = InvitationMessage(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                services=[TestConfig.test_target_did],
                requests_attach=[],
            )

            oob_record_retrieve_by_id.return_value = mock.MagicMock(
                state=OobRecord.STATE_ACCEPTED
            )

            result = await self.manager.receive_invitation(
                oob_invitation, use_existing_connection=True
            )

            oob_mgr_find_existing_conn.assert_called_once()
            assert result.state == OobRecord.STATE_ACCEPTED
            oob_record_save.assert_called_once_with(ANY, reason="Storing reuse msg data")

    async def test_receive_invitation_handshake_reuse(self):
        self.profile.context.update_settings({"public_invites": True})

        test_exist_conn = ConnRecord(
            connection_id="connection_id",
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-0123-4567-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )

        with (
            mock.patch.object(
                test_module.OutOfBandManager,
                "_handle_handshake_reuse",
                mock.CoroutineMock(),
            ) as handle_handshake_reuse,
            mock.patch.object(
                test_module.OutOfBandManager,
                "_perform_handshake",
                mock.CoroutineMock(),
            ) as perform_handshake,
            mock.patch.object(
                ConnRecord,
                "find_existing_connection",
                mock.CoroutineMock(return_value=test_exist_conn),
            ),
        ):
            oob_invitation = InvitationMessage(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                services=[TestConfig.test_target_did],
                requests_attach=[],
            )

            handle_handshake_reuse.return_value = mock.MagicMock(
                state=OobRecord.STATE_ACCEPTED
            )

            result = await self.manager.receive_invitation(
                oob_invitation, use_existing_connection=True
            )

            perform_handshake.assert_not_called()
            handle_handshake_reuse.assert_called_once_with(
                ANY, test_exist_conn, oob_invitation._version
            )

            assert result.state == OobRecord.STATE_ACCEPTED

    async def test_receive_invitation_handshake_reuse_failed(self):
        self.profile.context.update_settings({"public_invites": True})

        test_exist_conn = ConnRecord(
            connection_id="connection_id",
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-0123-4567-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )

        with (
            mock.patch.object(
                test_module.OutOfBandManager,
                "_handle_handshake_reuse",
                mock.CoroutineMock(),
            ) as handle_handshake_reuse,
            mock.patch.object(
                test_module.OutOfBandManager,
                "_perform_handshake",
                mock.CoroutineMock(),
            ) as perform_handshake,
            mock.patch.object(
                ConnRecord,
                "find_existing_connection",
                mock.CoroutineMock(return_value=test_exist_conn),
            ),
            mock.patch.object(
                ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=test_exist_conn),
            ),
        ):
            oob_invitation = InvitationMessage(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                services=[TestConfig.test_target_did],
                requests_attach=[],
            )

            mock_oob = mock.MagicMock(
                delete_record=mock.CoroutineMock(),
                emit_event=mock.CoroutineMock(),
            )
            perform_handshake.return_value = mock_oob

            handle_handshake_reuse.return_value = mock.MagicMock(
                state=OobRecord.STATE_NOT_ACCEPTED
            )

            result = await self.manager.receive_invitation(
                oob_invitation,
                use_existing_connection=True,
                alias="alias",
                auto_accept=True,
                mediation_id="mediation_id",
            )

            handle_handshake_reuse.assert_called_once_with(
                ANY, test_exist_conn, oob_invitation._version
            )
            perform_handshake.assert_called_once_with(
                oob_record=ANY,
                alias="alias",
                auto_accept=True,
                mediation_id="mediation_id",
                service_accept=None,
            )

            assert mock_oob.state == OobRecord.STATE_DONE
            assert result is mock_oob
            mock_oob.emit_event.assert_called_once()
            mock_oob.delete_record.assert_called_once()

    async def test_receive_invitation_services_with_service_did(self):
        self.profile.context.update_settings({"public_invites": True})

        mock_conn = mock.MagicMock(connection_id="dummy")

        with (
            mock.patch.object(test_module, "DIDXManager", autospec=True) as didx_mgr_cls,
            mock.patch.object(
                ConnRecord,
                "retrieve_by_id",
                mock.CoroutineMock(return_value=mock_conn),
            ),
        ):
            didx_mgr_cls.return_value = mock.MagicMock(
                receive_invitation=mock.CoroutineMock(return_value=mock_conn)
            )
            oob_invitation = InvitationMessage(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                services=[TestConfig.test_did],
                requests_attach=[],
            )

            invitation_record = await self.manager.receive_invitation(oob_invitation)
            assert invitation_record.invitation.services

    async def test_request_attach_oob_message_processor_connectionless(self):
        requests_attach: List[AttachDecorator] = [
            AttachDecorator.deserialize(deepcopy(TestConfig.req_attach_v1))
        ]

        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.handle_message = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(
            OobMessageProcessor, mock_oob_processor
        )

        mock_service_decorator = ServiceDecorator(
            endpoint=self.test_endpoint, recipient_keys=[self.test_verkey]
        )

        with (
            mock.patch.object(
                AskarWallet,
                "create_signing_key",
                mock.CoroutineMock(),
            ) as mock_create_signing_key,
            mock.patch.object(
                OutOfBandManager,
                "_service_decorator_from_service",
                mock.CoroutineMock(),
            ) as mock_service_decorator_from_service,
        ):
            mock_create_signing_key.return_value = KeyInfo(
                verkey="H3C2AVvLMv6gmMNam3uVAjZpfkcJCwDwnZn6z3wXmqPV",
                metadata={},
                key_type=ED25519,
            )
            mock_service_decorator_from_service.return_value = mock_service_decorator
            oob_invitation = InvitationMessage(
                handshake_protocols=[],
                services=[self.test_service],
                requests_attach=requests_attach,
            )

            oob_record = await self.manager.receive_invitation(
                oob_invitation, use_existing_connection=True
            )

            assert (
                oob_record.our_recipient_key
                == "H3C2AVvLMv6gmMNam3uVAjZpfkcJCwDwnZn6z3wXmqPV"
            )
            assert oob_record.our_service
            assert oob_record.state == OobRecord.STATE_PREPARE_RESPONSE

            mock_create_signing_key.assert_called_once_with(ED25519)
            mock_oob_processor.handle_message.assert_called_once_with(
                self.profile,
                [attachment.content for attachment in requests_attach],
                oob_record=oob_record,
                their_service=mock_service_decorator,
            )

    async def test_request_attach_oob_message_processor_connection(self):
        self.profile.context.update_settings({"public_invites": True})
        test_exist_conn = ConnRecord(
            connection_id="a-connection-id",
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-0123-4567-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
            state=ConnRecord.State.COMPLETED.rfc160,
        )

        requests_attach: List[AttachDecorator] = [
            AttachDecorator.deserialize(deepcopy(TestConfig.req_attach_v1))
        ]

        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.handle_message = mock.CoroutineMock()
        self.profile.context.injector.bind_instance(
            OobMessageProcessor, mock_oob_processor
        )

        with mock.patch.object(
            ConnRecord,
            "find_existing_connection",
            mock.CoroutineMock(),
        ) as oob_mgr_find_existing_conn:
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            oob_invitation = InvitationMessage(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                services=[TestConfig.test_target_did],
                requests_attach=requests_attach,
            )

            oob_record = await self.manager.receive_invitation(
                oob_invitation, use_existing_connection=True
            )

            mock_oob_processor.handle_message.assert_called_once_with(
                self.profile,
                [attachment.content for attachment in requests_attach],
                oob_record=oob_record,
                their_service=None,
            )

    async def test_request_attach_wait_for_conn_rec_active(self):
        async with self.profile.session() as session:
            self.profile.context.update_settings({"public_invites": True})
            test_exist_conn = ConnRecord(
                my_did=TestConfig.test_did,
                their_did=TestConfig.test_target_did,
                their_public_did=TestConfig.test_target_did,
                invitation_msg_id="12345678-0123-4567-1234-567812345678",
                their_role=ConnRecord.Role.REQUESTER,
            )

            with (
                mock.patch.object(
                    OutOfBandManager, "_wait_for_conn_rec_active"
                ) as mock_wait_for_conn_rec_active,
                mock.patch.object(
                    ConnRecord,
                    "find_existing_connection",
                    mock.CoroutineMock(),
                ) as oob_mgr_find_existing_conn,
            ):
                oob_mgr_find_existing_conn.return_value = test_exist_conn
                mock_wait_for_conn_rec_active.return_value = None
                oob_invitation = InvitationMessage(
                    handshake_protocols=[
                        pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                    ],
                    services=[TestConfig.test_target_did],
                    requests_attach=[
                        AttachDecorator.deserialize(deepcopy(TestConfig.req_attach_v1))
                    ],
                )

                with self.assertRaises(test_module.OutOfBandManagerError) as err:
                    oob_record = await self.manager.receive_invitation(
                        oob_invitation, use_existing_connection=True
                    )
                assert (
                    "Connection not ready to process attach message for connection_id:"
                    in err.exception.message
                )

    async def test_service_decorator_from_service_did(self):
        did = "did:sov:something"

        self.manager.resolve_invitation = mock.CoroutineMock()
        self.manager.resolve_invitation.return_value = (
            TestConfig.test_endpoint,
            [TestConfig.test_verkey],
            self.test_mediator_routing_keys,
        )

        service = await self.manager._service_decorator_from_service(did)

        assert service.endpoint == TestConfig.test_endpoint
        assert service.recipient_keys == [TestConfig.test_verkey]
        assert service.routing_keys == self.test_mediator_routing_keys

    async def test_service_decorator_from_service_object(self):
        oob_service = OobService(
            service_endpoint=TestConfig.test_endpoint,
            recipient_keys=[
                DIDKey.from_public_key_b58(TestConfig.test_verkey, ED25519).did
            ],
            routing_keys=[
                DIDKey.from_public_key_b58(verkey, ED25519).did
                for verkey in self.test_mediator_routing_keys
            ],
        )
        service = await self.manager._service_decorator_from_service(oob_service)

        assert service.endpoint == TestConfig.test_endpoint
        assert service.recipient_keys == [TestConfig.test_verkey]
        assert service.routing_keys == self.test_mediator_routing_keys

    async def test_service_decorator_from_service_str_empty_endpoint(self):
        did = "did:sov:something"

        self.manager.resolve_invitation = mock.CoroutineMock()
        self.manager.resolve_invitation.return_value = (
            "",  # empty endpoint
            [TestConfig.test_verkey],
            self.test_mediator_routing_keys,
        )

        service = await self.manager._service_decorator_from_service(did)

        assert service is None

    async def test_service_decorator_from_service_str_none_endpoint(self):
        did = "did:sov:something"

        self.manager.resolve_invitation = mock.CoroutineMock()
        self.manager.resolve_invitation.return_value = (
            None,  # None endpoint
            [TestConfig.test_verkey],
            self.test_mediator_routing_keys,
        )

        service = await self.manager._service_decorator_from_service(did)

        assert service is None

    async def test_service_decorator_from_service_object_empty_endpoint(self):
        oob_service = OobService(
            service_endpoint="",  # empty endpoint
            recipient_keys=[
                DIDKey.from_public_key_b58(TestConfig.test_verkey, ED25519).did
            ],
            routing_keys=[
                DIDKey.from_public_key_b58(verkey, ED25519).did
                for verkey in self.test_mediator_routing_keys
            ],
        )
        service = await self.manager._service_decorator_from_service(oob_service)

        assert service is None

    async def test_service_decorator_from_service_object_none_endpoint(self):
        oob_service = OobService(
            service_endpoint=None,  # None endpoint
            recipient_keys=[
                DIDKey.from_public_key_b58(TestConfig.test_verkey, ED25519).did
            ],
            routing_keys=[
                DIDKey.from_public_key_b58(verkey, ED25519).did
                for verkey in self.test_mediator_routing_keys
            ],
        )
        service = await self.manager._service_decorator_from_service(oob_service)

        assert service is None

    async def test_service_decorator_from_service_invalid_type(self):
        invalid_type = 123  # integer, for example

        service = await self.manager._service_decorator_from_service(invalid_type)

        assert service is None

    async def test_delete_stale_connection_by_invitation(self):
        current_datetime = datetime_now()
        older_datetime = current_datetime - timedelta(hours=4)
        records = [
            ConnRecord(
                my_did=self.test_did,
                their_did="FBmi5JLf5g58kDnNXMy4QM",
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                state=ConnRecord.State.INVITATION.rfc160,
                invitation_key="dummy2",
                invitation_mode="once",
                invitation_msg_id="test123",
                updated_at=datetime_to_str(older_datetime),
            )
        ]
        with (
            mock.patch.object(
                ConnRecord, "query", mock.CoroutineMock()
            ) as mock_connrecord_query,
            mock.patch.object(
                ConnRecord, "delete_record", mock.CoroutineMock()
            ) as mock_connrecord_delete,
        ):
            mock_connrecord_query.return_value = records
            await self.manager.delete_stale_connection_by_invitation("test123")
            mock_connrecord_delete.assert_called_once()

    async def test_delete_conn_and_oob_record_invitation(self):
        invitation = InvitationMessage()
        oob_records = [
            OobRecord(
                invitation=invitation,
                invi_msg_id=invitation._id,
                role=OobRecord.ROLE_RECEIVER,
                connection_id=self.test_conn_rec.connection_id,
                state=OobRecord.STATE_INITIAL,
            )
        ]
        conn_records = [
            ConnRecord(
                my_did=self.test_did,
                their_did="FBmi5JLf5g58kDnNXMy4QM",
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                state=ConnRecord.State.INVITATION.rfc160,
                invitation_key="dummy2",
                invitation_mode="once",
                invitation_msg_id="test123",
            )
        ]
        with (
            mock.patch.object(
                ConnRecord, "query", mock.CoroutineMock()
            ) as mock_connrecord_query,
            mock.patch.object(
                ConnRecord, "delete_record", mock.CoroutineMock()
            ) as mock_connrecord_delete,
            mock.patch.object(
                OobRecord, "query", mock.CoroutineMock()
            ) as mock_oobrecord_query,
            mock.patch.object(
                OobRecord, "delete_record", mock.CoroutineMock()
            ) as mock_oobrecord_delete,
        ):
            mock_connrecord_query.return_value = conn_records
            mock_oobrecord_query.return_value = oob_records
            await self.manager.delete_conn_and_oob_record_invitation("test123")
            mock_connrecord_delete.assert_called_once()
            mock_oobrecord_delete.assert_called_once()

    async def test_fetch_invitation_succeeds(self):
        self.profile.context.update_settings({"public_invites": True})

        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did,
                TestConfig.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=[HSProto.RFC23],
            )

            invi_rec = await self.manager.fetch_oob_invitation_record_by_id(
                oob_id=invi_rec.oob_id
            )

            assert invi_rec.invitation._type == DIDCommPrefix.qualify_current(
                self.TEST_INVI_MESSAGE_TYPE
            )
            assert not invi_rec.invitation.requests_attach
            assert (
                DIDCommPrefix.qualify_current(HSProto.RFC23.name)
                in invi_rec.invitation.handshake_protocols
            )
            assert invi_rec.invitation.services == [f"did:sov:{TestConfig.test_did}"]



"""Rotate ack message."""

from marshmallow import EXCLUDE, ValidationError, pre_dump

from ....notification.v1_0.messages.ack import V10Ack, V10AckSchema
from ..message_types import ACK, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.ack_handler.RotateAckHandler"


class RotateAck(V10Ack):
    """Rotate ack."""

    class Meta:
        """Rotate ack metadata."""

        handler_class = HANDLER_CLASS
        schema_class = "RotateAckSchema"
        message_type = ACK

    def __init__(self, **kwargs):
        """Initialize rotate ack object."""
        super().__init__(**kwargs)


class RotateAckSchema(V10AckSchema):
    """Rotate ack schema."""

    class Meta:
        """Schema metadata."""

        model_class = RotateAck
        unknown = EXCLUDE

    @pre_dump
    def check_thread_deco(self, obj, **kwargs):
        """Thread decorator, and its thid, are mandatory."""
        if not obj._decorators.to_dict().get("~thread", {}).keys() >= {"thid"}:
            raise ValidationError("Missing required field(s) in thread decorator")
        return obj



"""Credential definition admin routes."""

import json

# from asyncio import ensure_future, shield
from asyncio import shield
from time import time

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields

from ...admin.decorators.auth import tenant_authentication
from ...admin.request_context import AdminRequestContext
from ...connections.models.conn_record import ConnRecord
from ...core.event_bus import Event, EventBus
from ...core.profile import Profile
from ...indy.issuer import IndyIssuer, IndyIssuerError
from ...indy.models.cred_def import CredentialDefinitionSchema
from ...ledger.base import BaseLedger
from ...ledger.error import LedgerError
from ...ledger.multiple_ledger.ledger_requests_executor import (
    GET_CRED_DEF,
    IndyLedgerRequestsExecutor,
)
from ...multitenant.base import BaseMultitenantManager
from ...protocols.endorse_transaction.v1_0.manager import (
    TransactionManager,
    TransactionManagerError,
)
from ...protocols.endorse_transaction.v1_0.models.transaction_record import (
    TransactionRecordSchema,
)
from ...protocols.endorse_transaction.v1_0.util import (
    get_endorser_connection_id,
    is_author_role,
)
from ...revocation.indy import IndyRevocation
from ...storage.base import BaseStorage, StorageRecord
from ...storage.error import StorageError, StorageNotFoundError
from ...utils.profiles import is_anoncreds_profile_raise_web_exception
from ..models.base import BaseModelError
from ..models.openapi import OpenAPISchema
from ..valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_REV_REG_SIZE_EXAMPLE,
    INDY_REV_REG_SIZE_VALIDATE,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    UUID4_EXAMPLE,
)
from .util import (
    CRED_DEF_SENT_RECORD_TYPE,
    CRED_DEF_TAGS,
    EVENT_LISTENER_PATTERN,
    CredDefQueryStringSchema,
    notify_cred_def_event,
)


class CredentialDefinitionSendRequestSchema(OpenAPISchema):
    """Request schema for schema send request."""

    schema_id = fields.Str(
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    support_revocation = fields.Boolean(
        required=False, metadata={"description": "Revocation supported flag"}
    )
    revocation_registry_size = fields.Int(
        required=False,
        validate=INDY_REV_REG_SIZE_VALIDATE,
        metadata={
            "description": "Revocation registry size",
            "strict": True,
            "example": INDY_REV_REG_SIZE_EXAMPLE,
        },
    )
    tag = fields.Str(
        required=False,
        dump_default="default",
        metadata={
            "description": "Credential definition identifier tag",
            "example": "default",
        },
    )


class CredentialDefinitionSendResultSchema(OpenAPISchema):
    """Result schema content for schema send request with auto-endorse."""

    credential_definition_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )


class TxnOrCredentialDefinitionSendResultSchema(OpenAPISchema):
    """Result schema for credential definition send request."""

    sent = fields.Nested(
        CredentialDefinitionSendResultSchema(),
        required=False,
        metadata={"definition": "Content sent"},
    )
    txn = fields.Nested(
        TransactionRecordSchema(),
        required=False,
        metadata={"description": "Credential definition transaction to endorse"},
    )


class CredentialDefinitionGetResultSchema(OpenAPISchema):
    """Result schema for schema get request."""

    credential_definition = fields.Nested(CredentialDefinitionSchema)


class CredentialDefinitionsCreatedResultSchema(OpenAPISchema):
    """Result schema for cred-defs-created request."""

    credential_definition_ids = fields.List(
        fields.Str(
            validate=INDY_CRED_DEF_ID_VALIDATE,
            metadata={
                "description": "Credential definition identifiers",
                "example": INDY_CRED_DEF_ID_EXAMPLE,
            },
        )
    )


class CredDefIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking cred def id."""

    cred_def_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )


class CreateCredDefTxnForEndorserOptionSchema(OpenAPISchema):
    """Class for user to input whether to create a transaction for endorser or not."""

    create_transaction_for_endorser = fields.Boolean(
        required=False,
        metadata={"description": "Create Transaction For Endorser's signature"},
    )


class CredDefConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


@docs(
    tags=["credential-definition"],
    summary="Sends a credential definition to the ledger",
)
@request_schema(CredentialDefinitionSendRequestSchema())
@querystring_schema(CreateCredDefTxnForEndorserOptionSchema())
@querystring_schema(CredDefConnIdMatchInfoSchema())
@response_schema(TxnOrCredentialDefinitionSendResultSchema(), 200, description="")
@tenant_authentication
async def credential_definitions_send_credential_definition(request: web.BaseRequest):
    """Request handler for sending a credential definition to the ledger.

    Args:
        request: aiohttp request object

    Returns:
        The credential definition identifier

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    outbound_handler = request["outbound_message_router"]

    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    write_ledger = not create_transaction_for_endorser
    endorser_did = None
    connection_id = request.query.get("conn_id")

    body = await request.json()

    schema_id = body.get("schema_id")
    support_revocation = bool(body.get("support_revocation"))
    tag = body.get("tag")
    rev_reg_size = body.get("revocation_registry_size")

    # Don't allow revocable cred def to be created without tails server base url
    if not profile.settings.get("tails_server_base_url") and support_revocation:
        raise web.HTTPBadRequest(
            reason="tails_server_base_url not configured. Can't create revocable credential definition."  # noqa: E501
        )

    tag_query = {"schema_id": schema_id}
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        found = await storage.find_all_records(
            type_filter=CRED_DEF_SENT_RECORD_TYPE,
            tag_query=tag_query,
        )
        if 0 < len(found):
            # need to check the 'tag' value
            for record in found:
                cred_def_id = record.value
                cred_def_id_parts = cred_def_id.split(":")
                if tag == cred_def_id_parts[4]:
                    raise web.HTTPBadRequest(
                        reason=f"Cred def for {schema_id} {tag} already exists"
                    )

    # check if we need to endorse
    if is_author_role(profile):
        # authors cannot write to the ledger
        write_ledger = False
        create_transaction_for_endorser = True
        if not connection_id:
            # author has not provided a connection id, so determine which to use
            connection_id = await get_endorser_connection_id(profile)
            if not connection_id:
                raise web.HTTPBadRequest(reason="No endorser connection found")

    if not write_ledger:
        try:
            async with profile.session() as session:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, connection_id
                )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except BaseModelError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        async with profile.session() as session:
            endorser_info = await connection_record.metadata_get(session, "endorser_info")
        if not endorser_info:
            raise web.HTTPForbidden(
                reason=(
                    "Endorser Info is not set up in "
                    "connection metadata for this connection record"
                )
            )
        if "endorser_did" not in endorser_info.keys():
            raise web.HTTPForbidden(
                reason=(
                    ' "endorser_did" is not set in "endorser_info"'
                    " in connection metadata for this connection record"
                )
            )
        endorser_did = endorser_info["endorser_did"]

    ledger = context.inject_or(BaseLedger)
    if not ledger:
        reason = "No ledger available"
        if not context.settings.get_value("wallet.type"):
            reason += ": missing wallet-type?"
        raise web.HTTPForbidden(reason=reason)

    issuer = context.inject(IndyIssuer)
    try:  # even if in wallet, send it and raise if erroneously so
        async with ledger:
            (cred_def_id, cred_def, novel) = await shield(
                ledger.create_and_send_credential_definition(
                    issuer,
                    schema_id,
                    signature_type=None,
                    tag=tag,
                    support_revocation=support_revocation,
                    write_ledger=write_ledger,
                    endorser_did=endorser_did,
                )
            )

    except (IndyIssuerError, LedgerError) as e:
        raise web.HTTPBadRequest(reason=e.message) from e

    issuer_did = cred_def_id.split(":")[0]
    meta_data = {
        "context": {
            "schema_id": schema_id,
            "cred_def_id": cred_def_id,
            "issuer_did": issuer_did,
            "support_revocation": support_revocation,
            "novel": novel,
            "tag": tag,
            "rev_reg_size": rev_reg_size,
        },
        "processing": {
            "create_pending_rev_reg": True,
        },
    }

    if not create_transaction_for_endorser:
        # Notify event
        meta_data["processing"]["auto_create_rev_reg"] = True
        await notify_cred_def_event(profile, cred_def_id, meta_data)

        return web.json_response(
            {
                "sent": {"credential_definition_id": cred_def_id},
                "credential_definition_id": cred_def_id,
            }
        )

    # If the transaction is for the endorser, but the schema has already been created,
    # then we send back the schema since the transaction will fail to be created.
    elif "signed_txn" not in cred_def:
        return web.json_response({"sent": {"credential_definition_id": cred_def_id}})
    else:
        meta_data["processing"]["auto_create_rev_reg"] = context.settings.get_value(
            "endorser.auto_create_rev_reg"
        )

        transaction_mgr = TransactionManager(profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=cred_def["signed_txn"],
                connection_id=connection_id,
                meta_data=meta_data,
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if context.settings.get_value("endorser.auto_request"):
            try:
                transaction, transaction_request = await transaction_mgr.create_request(
                    transaction=transaction,
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=connection_id)

        return web.json_response(
            {
                "sent": {"credential_definition_id": cred_def_id},
                "txn": transaction.serialize(),
            }
        )


@docs(
    tags=["credential-definition"],
    summary="Search for matching credential definitions that agent originated",
)
@querystring_schema(CredDefQueryStringSchema())
@response_schema(CredentialDefinitionsCreatedResultSchema(), 200, description="")
@tenant_authentication
async def credential_definitions_created(request: web.BaseRequest):
    """Request handler for retrieving credential definitions that current agent created.

    Args:
        request: aiohttp request object

    Returns:
        The identifiers of matching credential definitions.

    """
    context: AdminRequestContext = request["context"]

    is_anoncreds_profile_raise_web_exception(context.profile)

    session = await context.session()
    storage = session.inject(BaseStorage)
    found = await storage.find_all_records(
        type_filter=CRED_DEF_SENT_RECORD_TYPE,
        tag_query={
            tag: request.query[tag] for tag in CRED_DEF_TAGS if tag in request.query
        },
    )

    return web.json_response(
        {"credential_definition_ids": [record.value for record in found]}
    )


@docs(
    tags=["credential-definition"],
    summary="Gets a credential definition from the ledger",
)
@match_info_schema(CredDefIdMatchInfoSchema())
@response_schema(CredentialDefinitionGetResultSchema(), 200, description="")
@tenant_authentication
async def credential_definitions_get_credential_definition(request: web.BaseRequest):
    """Request handler for getting a credential definition from the ledger.

    Args:
        request: aiohttp request object

    Returns:
        The credential definition details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    cred_def_id = request.match_info["cred_def_id"]

    async with profile.session() as session:
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)
    ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
        cred_def_id,
        txn_record_type=GET_CRED_DEF,
    )
    if not ledger:
        reason = "No ledger available"
        if not context.settings.get_value("wallet.type"):
            reason += ": missing wallet-type?"
        raise web.HTTPForbidden(reason=reason)

    async with ledger:
        cred_def = await ledger.get_credential_definition(cred_def_id)

    if ledger_id:
        return web.json_response(
            {"ledger_id": ledger_id, "credential_definition": cred_def}
        )
    else:
        return web.json_response({"credential_definition": cred_def})


@docs(
    tags=["credential-definition"],
    summary="Writes a credential definition non-secret record to the wallet",
)
@match_info_schema(CredDefIdMatchInfoSchema())
@response_schema(CredentialDefinitionGetResultSchema(), 200, description="")
@tenant_authentication
async def credential_definitions_fix_cred_def_wallet_record(request: web.BaseRequest):
    """Request handler for fixing a credential definition wallet non-secret record.

    Args:
        request: aiohttp request object

    Returns:
        The credential definition details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    cred_def_id = request.match_info["cred_def_id"]

    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)
    ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
        cred_def_id,
        txn_record_type=GET_CRED_DEF,
    )
    if not ledger:
        reason = "No ledger available"
        if not context.settings.get_value("wallet.type"):
            reason += ": missing wallet-type?"
        raise web.HTTPForbidden(reason=reason)

    async with ledger:
        cred_def = await ledger.get_credential_definition(cred_def_id)
        cred_def_id_parts = cred_def_id.split(":")
        schema_seq_no = cred_def_id_parts[3]
        schema_response = await ledger.get_schema(schema_seq_no)
        schema_id = schema_response["id"]
        iss_did = cred_def_id_parts[0]

        # check if the record exists, if not add it
        found = await storage.find_all_records(
            type_filter=CRED_DEF_SENT_RECORD_TYPE,
            tag_query={
                "cred_def_id": cred_def_id,
            },
        )
        if 0 == len(found):
            await add_cred_def_non_secrets_record(
                session.profile, schema_id, iss_did, cred_def_id
            )

    if ledger_id:
        return web.json_response(
            {"ledger_id": ledger_id, "credential_definition": cred_def}
        )
    else:
        return web.json_response({"credential_definition": cred_def})


def register_events(event_bus: EventBus):
    """Subscribe to any events we need to support."""
    event_bus.subscribe(EVENT_LISTENER_PATTERN, on_cred_def_event)


async def on_cred_def_event(profile: Profile, event: Event):
    """Handle any events we need to support."""
    schema_id = event.payload["context"]["schema_id"]
    cred_def_id = event.payload["context"]["cred_def_id"]
    issuer_did = event.payload["context"]["issuer_did"]

    # after the ledger record is written, write the wallet non-secrets record
    await add_cred_def_non_secrets_record(profile, schema_id, issuer_did, cred_def_id)

    # check if we need to kick off the revocation registry setup
    meta_data = event.payload
    support_revocation = meta_data["context"]["support_revocation"]
    novel = meta_data["context"]["novel"]
    rev_reg_size = (
        meta_data["context"].get("rev_reg_size", None) if support_revocation else None
    )
    auto_create_rev_reg = meta_data["processing"].get("auto_create_rev_reg", False)
    create_pending_rev_reg = meta_data["processing"].get("create_pending_rev_reg", False)
    endorser_connection_id = (
        meta_data["endorser"].get("connection_id", None)
        if "endorser" in meta_data
        else None
    )
    if support_revocation and novel and auto_create_rev_reg:
        # this kicks off the revocation registry creation process, which is 3 steps:
        # 1 - create revocation registry (ledger transaction may require endorsement)
        # 2 - upload tails file
        # 3 - create revocation entry (ledger transaction may require endorsement)
        # For a cred def we also automatically create a second "pending" revocation
        # registry, so when the first one fills up we can continue to issue credentials
        # without a delay
        revoc = IndyRevocation(profile)
        await revoc.init_issuer_registry(
            cred_def_id,
            rev_reg_size,
            create_pending_rev_reg=create_pending_rev_reg,
            endorser_connection_id=endorser_connection_id,
        )


async def add_cred_def_non_secrets_record(
    profile: Profile, schema_id: str, issuer_did: str, credential_definition_id: str
):
    """Write the wallet non-secrets record for cred def (already written to the ledger).

    Note that the cred def private key signing informtion must already exist in the
    wallet.

    Args:
        profile: The active profile
        schema_id: The schema id (or stringified sequence number)
        issuer_did: The DID of the issuer
        credential_definition_id: The credential definition id

    """
    schema_id_parts = schema_id.split(":")
    cred_def_tags = {
        "schema_id": schema_id,
        "schema_issuer_did": schema_id_parts[0],
        "schema_name": schema_id_parts[-2],
        "schema_version": schema_id_parts[-1],
        "issuer_did": issuer_did,
        "cred_def_id": credential_definition_id,
        "epoch": str(int(time())),
    }
    record = StorageRecord(
        CRED_DEF_SENT_RECORD_TYPE, credential_definition_id, cred_def_tags
    )
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        await storage.add_record(record)


async def register(app: web.Application):
    """Register routes."""
    app.add_routes(
        [
            web.post(
                "/credential-definitions",
                credential_definitions_send_credential_definition,
            ),
            web.get(
                "/credential-definitions/created",
                credential_definitions_created,
                allow_head=False,
            ),
            web.get(
                "/credential-definitions/{cred_def_id}",
                credential_definitions_get_credential_definition,
                allow_head=False,
            ),
            web.post(
                "/credential-definitions/{cred_def_id}/write_record",
                credential_definitions_fix_cred_def_wallet_record,
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "credential-definition",
            "description": "Credential definition operations",
            "externalDocs": {
                "description": "Specification",
                "url": (
                    "https://github.com/hyperledger/indy-node/blob/main/"
                    "design/anoncreds.md#cred_def"
                ),
            },
        }
    )



"""EcdsaSecp256r1Signature2019 suite."""

from datetime import datetime
from typing import List, Optional, Union

from ....utils.multiformats import multibase
from ..crypto import _KeyPair as KeyPair
from ..document_loader import DocumentLoaderMethod
from ..error import LinkedDataProofException
from .linked_data_signature import LinkedDataSignature


class EcdsaSecp256r1Signature2019(LinkedDataSignature):
    """EcdsaSecp256r1Signature2019 suite."""

    signature_type = "EcdsaSecp256r1Signature2019"

    def __init__(
        self,
        *,
        key_pair: KeyPair,
        proof: Optional[dict] = None,
        verification_method: Optional[str] = None,
        date: Union[datetime, str, None] = None,
    ):
        """Create new EcdsaSecp256r1Signature2019 instance.

        Args:
            key_pair (KeyPair): Key pair to use. Must provide ES256 signatures (P256)
            proof (dict, optional): A JSON-LD document with options to use for the
                `proof` node (e.g. any other custom fields can be provided here
                using a context different from security-v2).
            verification_method (str, optional): A key id URL to the paired public key.
            date (datetime, optional): Signing date to use.
        """
        super().__init__(
            verification_method=verification_method,
            proof=proof,
            date=date,
        )
        self.key_pair = key_pair

    async def sign(self, *, verify_data: bytes, proof: dict) -> dict:
        """Sign the data and add it to the proof.

        Args:
            verify_data (List[bytes]): The data to sign.
            proof (dict): The proof to add the signature to

        Returns:
            dict: The proof object with the added signature

        """
        signature = await self.key_pair.sign(verify_data)

        proof["proofValue"] = multibase.encode(signature, "base58btc")

        return proof

    async def verify_signature(
        self,
        *,
        verify_data: List[bytes],
        verification_method: dict,
        document: dict,
        proof: dict,
        document_loader: DocumentLoaderMethod,
    ) -> bool:
        """Verify the data against the proof.

        Args:
            verify_data (bytes): The data to check
            verification_method (dict): The verification method to use.
            document (dict): The document the verify data is derived for as extra context
            proof (dict): The proof to check
            document_loader (DocumentLoader): Document loader used for resolving

        Returns:
            bool: Whether the signature is valid for the data

        """

        if not (isinstance(proof.get("proofValue"), str)):
            raise LinkedDataProofException(
                'The proof does not contain a valid "proofValue" property.'
            )

        signature = multibase.decode(proof["proofValue"])

        # If the key pair has no public key yet, create a new key pair
        # from the verification method. We don't want to overwrite data
        # on the original key pair
        key_pair = self.key_pair
        if not key_pair.has_public_key:
            key_pair = key_pair.from_verification_method(verification_method)

        return await key_pair.verify(verify_data, signature)



"""Base key pair class."""

from abc import ABC, abstractmethod
from typing import List, Optional, Union


class KeyPair(ABC):
    """Base key pair class."""

    @abstractmethod
    async def sign(self, message: Union[List[bytes], bytes]) -> bytes:
        """Sign message(s) using key pair."""

    @abstractmethod
    async def verify(self, message: Union[List[bytes], bytes], signature: bytes) -> bool:
        """Verify message(s) against signature using key pair."""

    @property
    @abstractmethod
    def has_public_key(self) -> bool:
        """Whether key pair has a public key.

        Public key is required for verification, but can be set dynamically
        in the verification process.
        """

    @property
    @abstractmethod
    def public_key(self) -> Optional[bytes]:
        """Getter for the public key bytes.

        Returns:
            bytes: The public key

        """

    @abstractmethod
    def from_verification_method(self, verification_method: dict) -> "KeyPair":
        """Create new key pair class based on the passed verification method."""



"""Basic message."""

from datetime import datetime
from typing import Optional, Union

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.util import datetime_now, datetime_to_str
from .....messaging.valid import ISO8601_DATETIME_EXAMPLE, ISO8601_DATETIME_VALIDATE
from ..message_types import BASIC_MESSAGE, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.basicmessage_handler.BasicMessageHandler"


class BasicMessage(AgentMessage):
    """Class defining the structure of a basic message."""

    class Meta:
        """Basic message metadata class."""

        handler_class = HANDLER_CLASS
        message_type = BASIC_MESSAGE
        schema_class = "BasicMessageSchema"

    def __init__(
        self,
        *,
        sent_time: Union[str, datetime] = None,
        content: Optional[str] = None,
        localization: Optional[str] = None,
        **kwargs,
    ):
        """Initialize basic message object.

        Args:
            sent_time: Time message was sent
            content: message content
            localization: localization
            kwargs: additional keyword arguments

        """
        super().__init__(**kwargs)
        if not sent_time:
            sent_time = datetime_now()
        if localization:
            self._decorators["l10n"] = localization
        self.sent_time = datetime_to_str(sent_time)
        self.content = content


class BasicMessageSchema(AgentMessageSchema):
    """Basic message schema class."""

    class Meta:
        """Basic message schema metadata."""

        model_class = BasicMessage
        unknown = EXCLUDE

    sent_time = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": (
                "Time message was sent, ISO8601 with space date/time separator"
            ),
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )
    content = fields.Str(
        required=True, metadata={"description": "Message content", "example": "Hello"}
    )



import asyncio
import json
from collections import OrderedDict
from copy import deepcopy
from unittest import IsolatedAsyncioTestCase

import pytest

from ....cache.base import BaseCache
from ....cache.in_memory import InMemoryCache
from ....ledger.base import BaseLedger
from ....messaging.responder import BaseResponder
from ....tests import mock
from ....utils.testing import create_test_profile
from ...error import LedgerError
from ...indy_vdr import IndyVdrLedger, IndyVdrLedgerPool
from ...merkel_validation.tests.test_data import GET_NYM_REPLY
from .. import indy_vdr_manager as test_module
from ..base_manager import MultipleLedgerManagerError
from ..indy_vdr_manager import MultiIndyVDRLedgerManager

GET_NYM_INDY_VDR_REPLY = {
    "data": {
        "dest": "Av63wJYM7xYR4AiygYq4c3",
        "identifier": "V4SGRU86Z58d6TV7PBUe6f",
        "role": "101",
        "seqNo": 17794,
        "txnTime": 1632262244,
        "verkey": "6QSduYdf8Bi6t8PfNm5vNomGWDtXhmMmTRzaciudBXYJ",
    },
    "dest": "Av63wJYM7xYR4AiygYq4c3",
    "identifier": "LibindyDid111111111111",
    "reqId": 1632267113185021500,
    "seqNo": 17794,
    "state_proof": {
        "multi_signature": {
            "participants": ["Node2", "Node3", "Node1"],
            "signature": "Qye7WDGrhwEpr2MUmQ2hhm8yWAsUG6gKKf4TXxrw7BybGA96HWXLXhnV5gm5HBQCb4sDXiirTKuyWgMDyfDxKewya9mZhkGXf5WzaADFuaoJkTeSywqqmsrfpcHc2e49eEyncpCxFzhJn6sius4jLgJ7MAfSeVGwyydeR1YsJb3Nm5",
            "value": {
                "ledger_id": 1,
                "pool_state_root_hash": "7siDH8Qanh82UviK4zjBSfLXcoCvLaeGkrByi1ow9Tsm",
                "state_root_hash": "GJq4XL4pJYnDGg3MJ64y3QnfuezxsuBEezk5GC5yaZPM",
                "timestamp": 1632266842,
                "txn_root_hash": "BTnnWQ7imcHSoMykHLeYZX5q8eGEHWdbUydQNA4RG8La",
            },
        },
        "proof_nodes": r"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",
        "root_hash": "GJq4XL4pJYnDGg3MJ64y3QnfuezxsuBEezk5GC5yaZPM",
    },
    "txnTime": 1632262244,
    "type": "105",
}


@pytest.mark.indy_vdr
class TestMultiIndyVDRLedgerManager(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(BaseCache, InMemoryCache())
        self.context = self.profile.context
        setattr(self.context, "profile", self.profile)
        self.responder = mock.CoroutineMock(send=mock.CoroutineMock())
        self.context.injector.bind_instance(BaseResponder, self.responder)
        self.production_ledger = OrderedDict()
        self.non_production_ledger = OrderedDict()
        test_prod_ledger = IndyVdrLedger(IndyVdrLedgerPool("test_prod_1"), self.profile)
        writable_ledgers = set()
        self.production_ledger["test_prod_1"] = test_prod_ledger
        self.production_ledger["test_prod_2"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_prod_2"), self.profile
        )
        self.non_production_ledger["test_non_prod_1"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_non_prod_1"), self.profile
        )
        self.non_production_ledger["test_non_prod_2"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_non_prod_2"), self.profile
        )
        writable_ledgers.add("test_prod_1")
        writable_ledgers.add("test_prod_2")
        self.manager = MultiIndyVDRLedgerManager(
            self.profile,
            production_ledgers=self.production_ledger,
            non_production_ledgers=self.non_production_ledger,
            writable_ledgers=writable_ledgers,
        )

    def test_get_endorser_info_for_ledger(self):
        writable_ledgers = set()
        writable_ledgers.add("test_prod_1")
        writable_ledgers.add("test_prod_2")

        endorser_info_map = {}
        endorser_info_map["test_prod_1"] = {
            "endorser_did": "test_public_did_1",
            "endorser_alias": "endorser_1",
        }
        endorser_info_map["test_prod_2"] = {
            "endorser_did": "test_public_did_2",
            "endorser_alias": "endorser_2",
        }
        manager = MultiIndyVDRLedgerManager(
            self.profile,
            production_ledgers=self.production_ledger,
            non_production_ledgers=self.non_production_ledger,
            writable_ledgers=writable_ledgers,
            endorser_map=endorser_info_map,
        )
        endorser_info_1 = manager.get_endorser_info_for_ledger("test_prod_1")
        assert endorser_info_1 == ("endorser_1", "test_public_did_1")

        endorser_info_2 = manager.get_endorser_info_for_ledger("test_prod_2")
        assert endorser_info_2 == ("endorser_2", "test_public_did_2")

    async def test_get_write_ledgers(self):
        ledger_ids = await self.manager.get_write_ledgers()
        assert "test_prod_1" in ledger_ids
        assert "test_prod_2" in ledger_ids

    async def test_get_write_ledger_from_base_ledger(self):
        ledger_id = await self.manager.get_ledger_id_by_ledger_pool_name("test_prod_1")
        assert ledger_id == "test_prod_1"

    async def test_set_profile_write_ledger(self):
        writable_ledgers = set()
        writable_ledgers.add("test_prod_1")
        writable_ledgers.add("test_prod_2")
        endorser_info_map = {}
        endorser_info_map["test_prod_2"] = {
            "endorser_did": "test_public_did_2",
            "endorser_alias": "endorser_2",
        }
        manager = MultiIndyVDRLedgerManager(
            self.profile,
            production_ledgers=self.production_ledger,
            non_production_ledgers=self.non_production_ledger,
            writable_ledgers=writable_ledgers,
            endorser_map=endorser_info_map,
        )
        self.profile = await create_test_profile()
        assert not self.profile.inject_or(BaseLedger)
        assert "test_prod_2" in manager.writable_ledgers
        new_write_ledger_id = await manager.set_profile_write_ledger(
            profile=self.profile, ledger_id="test_prod_2"
        )
        assert new_write_ledger_id == "test_prod_2"
        new_write_ledger = self.profile.inject_or(BaseLedger)
        assert new_write_ledger.pool_name == "test_prod_2"

    async def test_set_profile_write_ledger_x(self):
        self.profile = await create_test_profile()
        with self.assertRaises(MultipleLedgerManagerError) as cm:
            await self.manager.set_profile_write_ledger(
                profile=self.profile, ledger_id="test_non_prod_1"
            )
        assert "is not write configurable" in str(cm.exception.message)

    async def test_get_ledger_inst_by_id(self):
        ledger_inst = await self.manager.get_ledger_inst_by_id("test_prod_2")
        assert ledger_inst
        ledger_inst = await self.manager.get_ledger_inst_by_id("test_non_prod_2")
        assert ledger_inst
        ledger_inst = await self.manager.get_ledger_inst_by_id("test_invalid")
        assert not ledger_inst

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_self_cert_a(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        with mock.patch.object(
            test_module.asyncio, "wait", mock.CoroutineMock()
        ) as mock_wait:
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = json.dumps(GET_NYM_INDY_VDR_REPLY)
            mock_wait.return_value = mock_submit.return_value
            (
                ledger_id,
                ledger_inst,
                is_self_certified,
            ) = await self.manager._get_ledger_by_did(
                "test_prod_1", "Av63wJYM7xYR4AiygYq4c3"
            )
            assert ledger_id == "test_prod_1"
            assert ledger_inst.pool.name == "test_prod_1"
            assert is_self_certified

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_self_cert_b(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        self.non_production_ledger = OrderedDict()
        self.non_production_ledger["test_non_prod_1"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_non_prod_1"), self.profile
        )
        self.non_production_ledger["test_non_prod_2"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_non_prod_2"), self.profile
        )
        self.manager = MultiIndyVDRLedgerManager(
            self.profile,
            non_production_ledgers=self.non_production_ledger,
        )
        with mock.patch.object(
            test_module.asyncio, "wait", mock.CoroutineMock()
        ) as mock_wait:
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = json.dumps(GET_NYM_INDY_VDR_REPLY)
            mock_wait.return_value = mock_submit.return_value
            (
                ledger_id,
                ledger_inst,
                is_self_certified,
            ) = await self.manager._get_ledger_by_did(
                "test_non_prod_1", "Av63wJYM7xYR4AiygYq4c3"
            )
            assert ledger_id == "test_non_prod_1"
            assert ledger_inst.pool.name == "test_non_prod_1"
            assert is_self_certified

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_not_self_cert(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        get_nym_reply = deepcopy(GET_NYM_INDY_VDR_REPLY)
        get_nym_reply["data"] = json.dumps(
            {
                "dest": "Av63wJYM7xYR4AiygYq4c3",
                "identifier": "V4SGRU86Z58d6TV7PBUe6f",
                "role": "101",
                "seqNo": 17794,
                "txnTime": 1632262244,
                "verkey": "ABUF7uxYTxZ6qYdZ4G9e1Gi",
            }
        )
        with (
            mock.patch.object(
                test_module.asyncio, "wait", mock.CoroutineMock()
            ) as mock_wait,
            mock.patch.object(
                test_module.SubTrie, "verify_spv_proof", mock.CoroutineMock()
            ) as mock_verify_spv_proof,
        ):
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = get_nym_reply
            mock_wait.return_value = mock_submit.return_value
            mock_verify_spv_proof.return_value = True
            (
                ledger_id,
                ledger_inst,
                is_self_certified,
            ) = await self.manager._get_ledger_by_did(
                "test_prod_1", "Av63wJYM7xYR4AiygYq4c3"
            )
            assert ledger_id == "test_prod_1"
            assert ledger_inst.pool.name == "test_prod_1"
            assert not is_self_certified

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_state_proof_not_valid(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        get_nym_reply = deepcopy(GET_NYM_REPLY)
        get_nym_reply["result"]["data"]["verkey"] = "ABUF7uxYTxZ6qYdZ4G9e1Gi"
        with mock.patch.object(
            test_module.asyncio, "wait", mock.CoroutineMock()
        ) as mock_wait:
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = json.dumps(get_nym_reply)
            mock_wait.return_value = mock_submit.return_value
            assert not await self.manager._get_ledger_by_did(
                "test_prod_1", "Av63wJYM7xYR4AiygYq4c3"
            )

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_no_data(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        get_nym_reply = deepcopy(GET_NYM_INDY_VDR_REPLY)
        get_nym_reply.pop("data")
        with mock.patch.object(
            test_module.asyncio, "wait", mock.CoroutineMock()
        ) as mock_wait:
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = get_nym_reply
            mock_wait.return_value = mock_submit.return_value
            assert not await self.manager._get_ledger_by_did(
                "test_prod_1", "Av63wJYM7xYR4AiygYq4c3"
            )

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_timeout(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        mock_build_get_nym_req.return_value = mock.MagicMock()
        mock_submit.side_effect = asyncio.TimeoutError
        assert not await self.manager._get_ledger_by_did(
            "test_prod_1", "Av63wJYM7xYR4AiygYq4c3"
        )

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_ledger_error(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        mock_build_get_nym_req.return_value = mock.MagicMock()
        mock_submit.side_effect = LedgerError
        assert not await self.manager._get_ledger_by_did(
            "test_prod_1", "Av63wJYM7xYR4AiygYq4c3"
        )

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_lookup_did_in_configured_ledgers_self_cert_prod(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        with mock.patch.object(
            test_module.asyncio, "wait", mock.CoroutineMock()
        ) as mock_wait:
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = json.dumps(GET_NYM_INDY_VDR_REPLY)
            mock_wait.return_value = mock_submit.return_value
            (
                ledger_id,
                ledger_inst,
            ) = await self.manager.lookup_did_in_configured_ledgers(
                "Av63wJYM7xYR4AiygYq4c3", cache_did=True
            )
            assert ledger_id == "test_prod_1"
            assert ledger_inst.pool.name == "test_prod_1"

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_not_self_cert_not_self_cert_prod(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        get_nym_reply = deepcopy(GET_NYM_INDY_VDR_REPLY)
        get_nym_reply["data"]["verkey"] = "ABUF7uxYTxZ6qYdZ4G9e1Gi"
        with (
            mock.patch.object(
                test_module.asyncio, "wait", mock.CoroutineMock()
            ) as mock_wait,
            mock.patch.object(
                test_module.SubTrie, "verify_spv_proof", mock.CoroutineMock()
            ) as mock_verify_spv_proof,
        ):
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = get_nym_reply
            mock_wait.return_value = mock_submit.return_value
            mock_verify_spv_proof.return_value = True
            (
                ledger_id,
                ledger_inst,
            ) = await self.manager.lookup_did_in_configured_ledgers(
                "Av63wJYM7xYR4AiygYq4c3", cache_did=True
            )
            assert ledger_id == "test_prod_1"
            assert ledger_inst.pool.name == "test_prod_1"

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_lookup_did_in_configured_ledgers_self_cert_non_prod(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        self.non_production_ledger = OrderedDict()
        self.non_production_ledger["test_non_prod_1"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_non_prod_1"), self.profile
        )
        self.non_production_ledger["test_non_prod_2"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_non_prod_2"), self.profile
        )
        self.manager = MultiIndyVDRLedgerManager(
            self.profile,
            non_production_ledgers=self.non_production_ledger,
        )
        with mock.patch.object(
            test_module.asyncio, "wait", mock.CoroutineMock()
        ) as mock_wait:
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = GET_NYM_INDY_VDR_REPLY
            mock_wait.return_value = mock_submit.return_value
            (
                ledger_id,
                ledger_inst,
            ) = await self.manager.lookup_did_in_configured_ledgers(
                "Av63wJYM7xYR4AiygYq4c3", cache_did=True
            )
            assert ledger_id == "test_non_prod_1"
            assert ledger_inst.pool.name == "test_non_prod_1"

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_get_ledger_by_did_not_self_cert_non_prod(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        self.non_production_ledger = OrderedDict()
        self.non_production_ledger["test_non_prod_1"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_non_prod_1"), self.profile
        )
        self.non_production_ledger["test_non_prod_2"] = IndyVdrLedger(
            IndyVdrLedgerPool("test_non_prod_2"), self.profile
        )
        self.manager = MultiIndyVDRLedgerManager(
            self.profile,
            non_production_ledgers=self.non_production_ledger,
        )
        get_nym_reply = deepcopy(GET_NYM_REPLY)
        get_nym_reply["result"]["data"]["verkey"] = "ABUF7uxYTxZ6qYdZ4G9e1Gi"
        with (
            mock.patch.object(
                test_module.asyncio, "wait", mock.CoroutineMock()
            ) as mock_wait,
            mock.patch.object(
                test_module.SubTrie, "verify_spv_proof", mock.CoroutineMock()
            ) as mock_verify_spv_proof,
        ):
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = get_nym_reply
            mock_wait.return_value = mock_submit.return_value
            mock_verify_spv_proof.return_value = True
            (
                ledger_id,
                ledger_inst,
            ) = await self.manager.lookup_did_in_configured_ledgers(
                "Av63wJYM7xYR4AiygYq4c3", cache_did=True
            )
            assert ledger_id == "test_non_prod_1"
            assert ledger_inst.pool.name == "test_non_prod_1"

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_lookup_did_in_configured_ledgers_x(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        with (
            mock.patch.object(
                test_module.asyncio, "wait", mock.CoroutineMock()
            ) as mock_wait,
            mock.patch.object(
                test_module.SubTrie, "verify_spv_proof", mock.CoroutineMock()
            ) as mock_verify_spv_proof,
        ):
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = GET_NYM_INDY_VDR_REPLY
            mock_wait.return_value = mock_submit.return_value
            mock_verify_spv_proof.return_value = False
            with self.assertRaises(MultipleLedgerManagerError):
                await self.manager.lookup_did_in_configured_ledgers(
                    "Av63wJYM7xYR4AiygYq4c3", cache_did=True
                )

    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_open")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool.context_close")
    @mock.patch("indy_vdr.ledger.build_get_nym_request")
    @mock.patch("acapy_agent.ledger.indy_vdr.IndyVdrLedger._submit")
    async def test_lookup_did_in_configured_ledgers_prod_not_cached(
        self, mock_submit, mock_build_get_nym_req, mock_close, mock_open
    ):
        with mock.patch.object(
            test_module.asyncio, "wait", mock.CoroutineMock()
        ) as mock_wait:
            mock_build_get_nym_req.return_value = mock.MagicMock()
            mock_submit.return_value = GET_NYM_INDY_VDR_REPLY
            mock_wait.return_value = mock_submit.return_value
            (
                ledger_id,
                ledger_inst,
            ) = await self.manager.lookup_did_in_configured_ledgers(
                "Av63wJYM7xYR4AiygYq4c3", cache_did=False
            )
            assert ledger_id == "test_prod_1"
            assert ledger_inst.pool.name == "test_prod_1"

    async def test_lookup_did_in_configured_ledgers_cached_prod_ledger(self):
        cache = InMemoryCache()
        await cache.set("did_ledger_id_resolver::Av63wJYM7xYR4AiygYq4c3", "test_prod_1")
        self.profile.context.injector.bind_instance(BaseCache, cache)
        (
            ledger_id,
            ledger_inst,
        ) = await self.manager.lookup_did_in_configured_ledgers(
            "Av63wJYM7xYR4AiygYq4c3", cache_did=True
        )
        assert ledger_id == "test_prod_1"
        assert ledger_inst.pool.name == "test_prod_1"

    async def test_lookup_did_in_configured_ledgers_cached_non_prod_ledger(self):
        cache = InMemoryCache()
        await cache.set(
            "did_ledger_id_resolver::Av63wJYM7xYR4AiygYq4c3", "test_non_prod_2", None
        )
        self.profile.context.injector.bind_instance(BaseCache, cache)
        (
            ledger_id,
            ledger_inst,
        ) = await self.manager.lookup_did_in_configured_ledgers(
            "Av63wJYM7xYR4AiygYq4c3", cache_did=True
        )
        assert ledger_id == "test_non_prod_2"
        assert ledger_inst.pool.name == "test_non_prod_2"

    async def test_lookup_did_in_configured_ledgers_cached_x(self):
        cache = InMemoryCache()
        await cache.set("did_ledger_id_resolver::Av63wJYM7xYR4AiygYq4c3", "invalid_id")
        self.profile.context.injector.bind_instance(BaseCache, cache)
        with self.assertRaises(MultipleLedgerManagerError):
            await self.manager.lookup_did_in_configured_ledgers(
                "Av63wJYM7xYR4AiygYq4c3", cache_did=True
            )

    async def test_get_production_ledgers(self):
        assert len(await self.manager.get_prod_ledgers()) == 2

    async def test_get_non_production_ledgers(self):
        assert len(await self.manager.get_nonprod_ledgers()) == 2



BBS_NESTED_VC_MATTR = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "id": "https://example.gov/credentials/3732",
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "issuer": "did:example:489398593",
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "credentialSubject": {
        "id": "did:example:489398593",
        "degree": {
            "type": "BachelorDegree",
            "name": "Bachelor of Science and Arts",
            "degreeType": "Underwater Basket Weaving",
        },
        "college": "Contoso University",
    },
}



"""Dependency related util methods."""

import sys


def is_ursa_bbs_signatures_module_installed():
    """Check whether ursa_bbs_signatures module is installed.

    Returns:
        bool: Whether ursa_bbs_signatures is installed.

    """
    try:
        # Check if already imported
        if "ursa_bbs_signatures" in sys.modules:
            return True

        # Try to import
        import ursa_bbs_signatures  # noqa: F401

        return True
    except ModuleNotFoundError:
        # Not installed if import went wrong
        return False


def assert_ursa_bbs_signatures_installed():
    """Assert ursa_bbs_signatures module is installed."""
    if not is_ursa_bbs_signatures_module_installed():
        raise Exception("ursa_bbs_signatures module not installed")



"""AnonCreds Registry."""

import logging
from typing import List, Optional, Sequence

from ..core.profile import Profile
from .base import (
    AnonCredsRegistrationError,
    AnonCredsResolutionError,
    BaseAnonCredsHandler,
    BaseAnonCredsRegistrar,
    BaseAnonCredsResolver,
)
from .models.credential_definition import CredDef, CredDefResult, GetCredDefResult
from .models.revocation import (
    GetRevListResult,
    GetRevRegDefResult,
    RevList,
    RevListResult,
    RevRegDef,
    RevRegDefResult,
)
from .models.schema import AnonCredsSchema, GetSchemaResult, SchemaResult
from .models.schema_info import AnonCredsSchemaInfo

LOGGER = logging.getLogger(__name__)


class AnonCredsRegistry:
    """AnonCredsRegistry."""

    def __init__(self, registries: Optional[List[BaseAnonCredsHandler]] = None):
        """Create DID Resolver."""
        self.resolvers = []
        self.registrars = []
        if registries:
            for registry in registries:
                self.register(registry)

    def register(self, registry: BaseAnonCredsHandler) -> None:
        """Register a new registry."""
        if isinstance(registry, BaseAnonCredsResolver):
            self.resolvers.append(registry)
        if isinstance(registry, BaseAnonCredsRegistrar):
            self.registrars.append(registry)

    async def _resolver_for_identifier(self, identifier: str) -> BaseAnonCredsResolver:
        resolvers = [
            resolver for resolver in self.resolvers if await resolver.supports(identifier)
        ]
        if len(resolvers) == 0:
            raise AnonCredsResolutionError(
                f"No resolver available for identifier {identifier}"
            )
        if len(resolvers) > 1:
            raise AnonCredsResolutionError(
                f"More than one resolver found for identifier {identifier}"
            )
        return resolvers[0]

    async def _registrar_for_identifier(self, identifier: str) -> BaseAnonCredsRegistrar:
        registrars = [
            registrar
            for registrar in self.registrars
            if await registrar.supports(identifier)
        ]
        if len(registrars) == 0:
            raise AnonCredsRegistrationError(
                f"No registrar available for identifier {identifier}"
            )

        if len(registrars) > 1:
            raise AnonCredsRegistrationError(
                f"More than one registrar found for identifier {identifier}"
            )
        return registrars[0]

    async def get_schema(self, profile: Profile, schema_id: str) -> GetSchemaResult:
        """Get a schema from the registry."""
        resolver = await self._resolver_for_identifier(schema_id)
        return await resolver.get_schema(profile, schema_id)

    async def register_schema(
        self,
        profile: Profile,
        schema: AnonCredsSchema,
        options: Optional[dict] = None,
    ) -> SchemaResult:
        """Register a schema on the registry."""
        registrar = await self._registrar_for_identifier(schema.issuer_id)
        return await registrar.register_schema(profile, schema, options)

    async def get_credential_definition(
        self, profile: Profile, credential_definition_id: str
    ) -> GetCredDefResult:
        """Get a credential definition from the registry."""
        resolver = await self._resolver_for_identifier(credential_definition_id)
        return await resolver.get_credential_definition(
            profile,
            credential_definition_id,
        )

    async def get_schema_info_by_id(
        self, profile: Profile, schema_id: str
    ) -> AnonCredsSchemaInfo:
        """Get a schema info from the registry."""
        resolver = await self._resolver_for_identifier(schema_id)
        return await resolver.get_schema_info_by_id(profile, schema_id)

    async def register_credential_definition(
        self,
        profile: Profile,
        schema: GetSchemaResult,
        credential_definition: CredDef,
        options: Optional[dict] = None,
    ) -> CredDefResult:
        """Register a credential definition on the registry."""
        registrar = await self._registrar_for_identifier(credential_definition.issuer_id)
        return await registrar.register_credential_definition(
            profile,
            schema,
            credential_definition,
            options,
        )

    async def get_revocation_registry_definition(
        self, profile: Profile, revocation_registry_id: str
    ) -> GetRevRegDefResult:
        """Get a revocation registry definition from the registry."""
        resolver = await self._resolver_for_identifier(revocation_registry_id)
        return await resolver.get_revocation_registry_definition(
            profile, revocation_registry_id
        )

    async def register_revocation_registry_definition(
        self,
        profile: Profile,
        revocation_registry_definition: RevRegDef,
        options: Optional[dict] = None,
    ) -> RevRegDefResult:
        """Register a revocation registry definition on the registry."""
        registrar = await self._registrar_for_identifier(
            revocation_registry_definition.issuer_id
        )
        return await registrar.register_revocation_registry_definition(
            profile, revocation_registry_definition, options
        )

    async def get_revocation_list(
        self,
        profile: Profile,
        rev_reg_def_id: str,
        timestamp_from: Optional[int] = 0,
        timestamp_to: Optional[int] = None,
    ) -> GetRevListResult:
        """Get a revocation list from the registry."""
        resolver = await self._resolver_for_identifier(rev_reg_def_id)
        return await resolver.get_revocation_list(
            profile, rev_reg_def_id, timestamp_from, timestamp_to
        )

    async def register_revocation_list(
        self,
        profile: Profile,
        rev_reg_def: RevRegDef,
        rev_list: RevList,
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Register a revocation list on the registry."""
        registrar = await self._registrar_for_identifier(rev_list.issuer_id)
        return await registrar.register_revocation_list(
            profile, rev_reg_def, rev_list, options
        )

    async def update_revocation_list(
        self,
        profile: Profile,
        rev_reg_def: RevRegDef,
        prev_list: RevList,
        curr_list: RevList,
        revoked: Sequence[int],
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Update a revocation list on the registry."""
        registrar = await self._registrar_for_identifier(prev_list.issuer_id)
        return await registrar.update_revocation_list(
            profile, rev_reg_def, prev_list, curr_list, revoked, options
        )



"""Classes for configuring the default injection context."""

import logging

from ..anoncreds.registry import AnonCredsRegistry
from ..cache.base import BaseCache
from ..cache.in_memory import InMemoryCache
from ..connections.base_manager import BaseConnectionManager
from ..core.event_bus import EventBus
from ..core.goal_code_registry import GoalCodeRegistry
from ..core.plugin_registry import PluginRegistry
from ..core.profile import Profile, ProfileManager, ProfileManagerProvider
from ..core.protocol_registry import ProtocolRegistry
from ..protocols.actionmenu.v1_0.base_service import BaseMenuService
from ..protocols.actionmenu.v1_0.driver_service import DriverMenuService
from ..protocols.introduction.v0_1.base_service import BaseIntroductionService
from ..protocols.introduction.v0_1.demo_service import DemoIntroductionService
from ..resolver.did_resolver import DIDResolver
from ..transport.wire_format import BaseWireFormat
from ..utils.stats import Collector
from ..wallet.default_verification_key_strategy import (
    BaseVerificationKeyStrategy,
    DefaultVerificationKeyStrategy,
)
from ..wallet.did_method import DIDMethods
from ..wallet.key_type import KeyTypes
from .base_context import ContextBuilder
from .injection_context import InjectionContext
from .provider import CachedProvider, ClassProvider

LOGGER = logging.getLogger(__name__)


class DefaultContextBuilder(ContextBuilder):
    """Default context builder."""

    async def build_context(self) -> InjectionContext:
        """Build the base injection context; set DIDComm prefix to emit."""
        LOGGER.debug("Building new injection context")

        context = InjectionContext(settings=self.settings)
        context.settings.set_default("default_label", "Aries Cloud Agent")

        if context.settings.get("timing.enabled"):
            timing_log = context.settings.get("timing.log_file")
            LOGGER.debug("Enabling timing collector with log file: %s", timing_log)
            collector = Collector(log_path=timing_log)
            context.injector.bind_instance(Collector, collector)

        # Shared in-memory cache
        context.injector.bind_instance(BaseCache, InMemoryCache())

        # Global protocol registry
        context.injector.bind_instance(ProtocolRegistry, ProtocolRegistry())

        # Global goal code registry
        context.injector.bind_instance(GoalCodeRegistry, GoalCodeRegistry())

        # Global event bus
        context.injector.bind_instance(EventBus, EventBus())

        # Global did resolver
        context.injector.bind_instance(DIDResolver, DIDResolver())
        context.injector.bind_instance(AnonCredsRegistry, AnonCredsRegistry())
        context.injector.bind_instance(DIDMethods, DIDMethods())
        context.injector.bind_instance(KeyTypes, KeyTypes())
        context.injector.bind_instance(
            BaseVerificationKeyStrategy, DefaultVerificationKeyStrategy()
        )

        # DIDComm Messaging
        if context.settings.get("experiment.didcomm_v2"):
            LOGGER.info("DIDComm v2 experimental mode enabled")
            from didcomm_messaging import CryptoService, PackagingService, RoutingService
            from didcomm_messaging.crypto.backend.askar import AskarCryptoService

            context.injector.bind_instance(CryptoService, AskarCryptoService())
            context.injector.bind_instance(PackagingService, PackagingService())
            context.injector.bind_instance(RoutingService, RoutingService())

        await self.bind_providers(context)
        await self.load_plugins(context)

        return context

    async def bind_providers(self, context: InjectionContext):
        """Bind various class providers."""
        LOGGER.debug("Begin binding providers to context")

        context.injector.bind_provider(ProfileManager, ProfileManagerProvider())

        # Register default pack format
        context.injector.bind_provider(
            BaseWireFormat,
            CachedProvider(
                ClassProvider("acapy_agent.transport.pack_format.PackWireFormat"),
            ),
        )

        # Allow action menu to be provided by driver
        context.injector.bind_instance(BaseMenuService, DriverMenuService(context))
        context.injector.bind_instance(BaseIntroductionService, DemoIntroductionService())

        # Allow BaseConnectionManager to be overridden
        context.injector.bind_provider(
            BaseConnectionManager,
            ClassProvider(BaseConnectionManager, ClassProvider.Inject(Profile)),
        )

    async def load_plugins(self, context: InjectionContext):
        """Set up plugin registry and load plugins."""

        LOGGER.debug("Initializing plugin registry")
        plugin_registry = PluginRegistry(
            blocklist=self.settings.get("blocked_plugins", [])
        )
        context.injector.bind_instance(PluginRegistry, plugin_registry)

        # Register standard protocol plugins
        if not self.settings.get("transport.disabled"):
            plugin_registry.register_package("acapy_agent.protocols")

        # Define core plugins
        core_plugins = [
            "acapy_agent.holder",
            "acapy_agent.ledger",
            "acapy_agent.connections",
            "acapy_agent.messaging.jsonld",
            "acapy_agent.resolver",
            "acapy_agent.settings",
            "acapy_agent.vc",
            "acapy_agent.vc.data_integrity",
            "acapy_agent.wallet",
            "acapy_agent.wallet.keys",
        ]

        did_management_plugins = [
            "acapy_agent.did.indy",
        ]

        default_plugins = core_plugins + did_management_plugins

        LOGGER.info("Registering default plugins")
        for plugin in default_plugins:
            plugin_registry.register_plugin(plugin)

        anoncreds_plugins = [
            "acapy_agent.anoncreds",
            "acapy_agent.anoncreds.default.did_web",
            "acapy_agent.anoncreds.default.legacy_indy",
            "acapy_agent.revocation_anoncreds",
        ]

        askar_plugins = [
            "acapy_agent.messaging.credential_definitions",
            "acapy_agent.messaging.schemas",
            "acapy_agent.revocation",
        ]

        def register_askar_plugins():
            LOGGER.info("Registering askar plugins")
            for plugin in askar_plugins:
                plugin_registry.register_plugin(plugin)

        def register_anoncreds_plugins():
            LOGGER.info("Registering anoncreds plugins")
            for plugin in anoncreds_plugins:
                plugin_registry.register_plugin(plugin)

        if context.settings.get("multitenant.enabled"):
            # Register both askar and anoncreds plugins for multitenancy
            register_askar_plugins()
            register_anoncreds_plugins()
        elif self.settings.get("wallet.type") == "askar-anoncreds":
            register_anoncreds_plugins()
        else:
            register_askar_plugins()

        if context.settings.get("multitenant.admin_enabled"):
            LOGGER.info("Registering multitenant admin API plugin")
            plugin_registry.register_plugin("acapy_agent.multitenant.admin")

        # Register external plugins
        for plugin_path in self.settings.get("external_plugins", []):
            LOGGER.debug("Registering external plugin: %s", plugin_path)
            plugin_registry.register_plugin(plugin_path)

        # Register message protocols
        await plugin_registry.init_context(context)
